#!/usr/bin/env python3
"""Topology components for FAUCET Mininet unit tests."""

from collections import namedtuple
import os
import pty
import select
import socket
import string
import shutil
import subprocess
import tempfile
import time

import netifaces

# pylint: disable=too-many-arguments

from mininet.log import output, warn, error
from mininet.topo import Topo
from mininet.node import Controller
from mininet.node import Host
from mininet.node import OVSSwitch
from mininet.link import TCIntf, Link

from clib import mininet_test_util


SWITCH_START_PORT = 5


class FaucetIntf(TCIntf):
    """TCIntf that doesn't complain unnecessarily"""

    def delete(self):
        """Ignore interface deletion failure;
        this is common after a veth pair has been deleted
        on the other side."""
        self.cmd("ip link del", self.name, "|| true")
        self.node.delIntf(self)
        self.link = None


class FaucetLink(Link):
    """Link using FaucetIntfs"""

    def __init__(
        self,
        node1,
        node2,
        port1=None,
        port2=None,
        intf_name1=None,
        intf_name2=None,
        addr1=None,
        addr2=None,
        **params,
    ):
        Link.__init__(
            self,
            node1,
            node2,
            port1=port1,
            port2=port2,
            intfName1=intf_name1,
            intfName2=intf_name2,
            cls1=FaucetIntf,
            cls2=FaucetIntf,
            addr1=addr1,
            addr2=addr2,
            params1=params,
            params2=params,
        )


class FaucetHost(Host):
    """Base Mininet Host class, for Mininet-based tests."""

    def __init__(self, *args, **kwargs):
        self.pid_files = []
        # Isolate these directories in mininet container from host filesystem
        kwargs["privateDirs"] = [
            ("/etc/", "/var/tmp/mininet/%(name)s/etc"),
            ("/var/log/", "/var/tmp/mininet/%(name)s/var/log"),
            ("/run/", "/var/tmp/mininet/%(name)s/run"),
            ("/var/run/", "/var/tmp/mininet/%(name)s/var/run"),
        ]
        super().__init__(*args, **kwargs)

    def mountPrivateDirs(self):
        "Mount private directories"
        for directory in self.privateDirs:
            if isinstance(directory, tuple):
                # mount given private directory
                privateDir = directory[1] % self.__dict__
                mountPoint = directory[0]
                self.cmd(f"mkdir -p {privateDir}")
                self.cmd(f"cp -aRT {mountPoint} {privateDir}")
                self.cmd(f"mount --bind {privateDir} {mountPoint}")
            else:
                # mount temporary filesystem on directory
                self.cmd(f"mkdir -p {directory}")
                self.cmd(f"mount -n -t tmpfs tmpfs {directory}")

    def unmountPrivateDirs(self):
        "Unmount private directories"
        for directory in self.privateDirs:
            if isinstance(directory, tuple):
                self.cmd(f"umount {directory[0]}")
                self.cmd(f"rm -r {directory[1]}")
            else:
                self.cmd(f"umount {directory}")

    def startShell(self, mnopts=None):
        "Start a shell process for running commands"
        if self.shell:
            error("%s: shell is already running\n" % self.name)
            return
        # mnexec: (c)lose descriptors, (d)etach from tty,
        # (p)rint pid, and run in (n)amespace
        opts = "-cd" if mnopts is None else mnopts
        if self.inNamespace:
            opts += "n"
        # bash -i: force interactive
        # -s: pass $* to shell, and make process easy to find in ps
        # prompt is set to sentinel chr( 127 )
        cmd = [
            "mnexec",
            opts,
            "env",
            "PS1=" + chr(127),
            "dash",
            "-is",
            "mininet:" + self.name,
        ]

        # Spawn a shell subprocess in a pseudo-tty, to disable buffering
        # in the subprocess and insulate it from signals (e.g. SIGINT)
        # received by the parent
        self.master, self.slave = pty.openpty()
        self.shell = self._popen(
            cmd, stdin=self.slave, stdout=self.slave, stderr=self.slave, close_fds=False
        )
        # XXX BL: This doesn't seem right, and we should also probably
        # close our files when we exit...
        self.stdin = os.fdopen(self.master, "r")
        self.stdout = self.stdin
        self.pid = self.shell.pid
        self.pollOut = select.poll()
        self.pollOut.register(self.stdout)
        # Maintain mapping between file descriptors and nodes
        # This is useful for monitoring multiple nodes
        # using select.poll()
        self.outToNode[self.stdout.fileno()] = self
        self.inToNode[self.stdin.fileno()] = self
        self.execed = False
        self.lastCmd = None
        self.lastPid = None
        self.readbuf = ""
        # Wait for prompt
        while True:
            data = self.read(1024)
            if data[-1] == chr(127):
                break
            self.pollOut.poll()
        self.waiting = False
        # +m: disable job control notification
        self.cmd("unset HISTFILE; stty -echo; set +m")

    def terminate(self):
        # If any 'dnsmasq' processes were started, terminate them now
        for pid_file in self.pid_files:
            with open(pid_file, "r", encoding="utf-8") as pf:
                for _, pid in enumerate(pf):
                    os.kill(int(pid), 15)
        super().terminate()

    def create_dnsmasq(self, tmpdir, iprange, router, vlan, interface=None):
        """Start dnsmasq instance inside dnsmasq namespace"""
        if interface is None:
            interface = self.defaultIntf()
        dhcp_leasefile = os.path.join(
            tmpdir, "nfv-dhcp-%s-%s-vlan%u.leases" % (self.name, iprange, vlan)
        )
        log_facility = os.path.join(
            tmpdir, "nfv-dhcp-%s-%s-vlan%u.log" % (self.name, iprange, vlan)
        )
        pid_file = os.path.join(
            tmpdir, "dnsmasq-%s-%s-vlan%u.pid" % (self.name, iprange, vlan)
        )
        self.pid_files.append(pid_file)
        cmd = "dnsmasq"
        opts = ""
        opts += " --dhcp-range=%s,255.255.255.0" % iprange
        opts += " --dhcp-sequential-ip"
        opts += " --dhcp-option=option:router,%s" % router
        opts += " --no-resolv --txt-record=does.it.work,yes"
        opts += " --bind-interfaces"
        opts += " --except-interface=lo"
        opts += " --interface=%s" % interface
        opts += " --dhcp-leasefile=%s" % dhcp_leasefile
        opts += " --log-facility=%s" % log_facility
        opts += " --pid-file=%s" % pid_file
        opts += " --conf-file="
        return self.cmd(cmd + opts)

    def run_dhclient(self, tmpdir, interface=None, timeout=10):
        """Run DHCLIENT to obtain ip address via DHCP"""
        if interface is None:
            interface = self.defaultIntf()
        cmd = "dhclient"
        opts = ""
        opts += " -1"
        opts += " -d"
        opts += " -pf %s/dhclient-%s.pid" % (tmpdir, self.name)
        opts += " -lf %s/dhclient-%s.leases" % (tmpdir, self.name)
        opts += " %s" % interface
        dhclient_cmd = cmd + opts
        return self.cmd(
            mininet_test_util.timeout_cmd(dhclient_cmd, timeout), verbose=True
        )

    def return_ip(self):
        """Return host IP as a string"""
        return self.cmd("hostname -I")

    def run_ip_batch(self, cmds):
        with tempfile.TemporaryDirectory() as tmpdir:
            cmd_filename = os.path.join(tmpdir, "ipcmds.txt")
            with open(cmd_filename, "w") as cmd_file:
                cmd_file.write("\n".join(cmds))
            result = self.cmd("ip -b %s" % cmd_filename)
            assert result == "", (cmds, result)


class VLANHost(FaucetHost):
    """Implementation of a Mininet host on a tagged VLAN."""

    intf_root_name = None

    vlans = None
    vlan_intfs = None

    def config(self, vlans=None, **params):  # pylint: disable=arguments-differ
        """Configure VLANHost according to (optional) parameters:

        vlans (list): List of VLAN IDs (for the VLANs the host is configured to have)
            for default interface
        vlan_intfs (dict): Dictionary of interface IP addresses keyed by VLAN indices
        """
        super_config = super().config(**params)
        if vlans is None:
            vlans = [100]
        self.vlans = vlans
        self.vlan_intfs = {}
        batch_cmds = []
        intf = self.defaultIntf()
        self.intf_root_name = intf.name
        unique_intfs = set()

        def _config_vlan(root_name, full_name, vlan_id):
            batch_cmds.extend(
                [
                    "link add link %s name %s type vlan id %s"
                    % (root_name, full_name, vlan_id),
                    "link set dev %s up" % full_name,
                ]
            )

        def _config_ip(config_full_name, config_ip):
            batch_cmds.extend(["addr add %s dev %s" % (config_ip, config_full_name)])

        if "vlan_intfs" in params:
            vlan_intfs = params.get("vlan_intfs", {})
            for vlan_id, ip_addr in vlan_intfs.items():
                if isinstance(vlan_id, tuple):
                    # Interface will take multiply VLAN tagged packets
                    intf_name = "%s" % intf.name
                    for vlan_i in vlan_id:
                        prev_name = intf_name
                        # Cannot have intf name tu0xy-eth0.VID1.VID2 as that takes up too many bytes
                        intf_name += ".%s" % vlan_i
                        if intf_name not in unique_intfs:
                            _config_vlan(prev_name, intf_name, vlans[vlan_i])
                            unique_intfs.add(intf_name)
                        self.nameToIntf[intf_name] = intf
                        self.vlan_intfs.setdefault(vlan_id, [])
                        self.vlan_intfs[vlan_id].append(intf_name)
                    _config_ip(intf_name, ip_addr)
                else:
                    intf_name = "%s.%s" % (intf, vlans[vlan_id])
                    _config_vlan(intf, intf_name, vlans[vlan_id])
                    _config_ip(intf_name, ip_addr)
                    self.nameToIntf[intf_name] = intf
                    self.vlan_intfs[vlan_id] = intf_name
        else:
            for vlan_id in vlans:
                intf_name = "%s.%s" % (intf, vlan_id)
                _config_vlan(intf, intf_name, vlan_id)
                _config_ip(intf_name, params["ip"])
                self.nameToIntf[intf_name] = intf
                self.vlan_intfs.setdefault(vlan_id, [])
                self.vlan_intfs[vlan_id].append(intf_name)
            intf.name = intf_name

        batch_cmds.extend(["addr flush dev %s" % self.intf_root_name])
        self.run_ip_batch(batch_cmds)
        return super_config


class FaucetSwitch(OVSSwitch):
    """Switch that will be used by all tests (netdev based OVS)."""

    clist = None

    controller_params = {
        "controller_burst_limit": 25,
        "controller_rate_limit": 100,
    }

    def __init__(self, name, **params):
        self.clist = []
        super().__init__(name=name, reconnectms=8000, **params)

    @staticmethod
    def _workaround(args):
        """Workarounds/hacks for errors resulting from
        cmd() calls within Mininet"""
        # Workaround: ignore ethtool errors on tap interfaces
        # This allows us to use tap tunnels as cables to switch ports,
        # for example to test against OvS in a VM.
        if (
            len(args) > 1
            and args[0] == "ethtool -K"
            and getattr(args[1], "name", "").startswith("tap")
        ):
            return True
        return False

    def cmd(self, *args, success=0, **kwargs):
        """Commands typically must succeed for proper switch operation,
        so we check the exit code of the last command in *args.
        success: desired exit code (or None to skip check)"""
        # pylint: disable=arguments-differ
        cmd_output = super().cmd(*args, **kwargs)
        exit_code = int(super().cmd("echo $?"))
        if success is not None and exit_code != success:
            msg = "%s exited with (%d):'%s'" % (args, exit_code, cmd_output)
            if self._workaround(args):
                warn("Ignoring:", msg, "\n")
            else:
                raise RuntimeError(msg)
        return cmd_output

    def attach(self, intf):
        "Attach an interface and set its port"
        super().attach(intf)
        # This should be done in Mininet, but we do it for now
        port = self.ports[intf]
        self.cmd("ovs-vsctl set Interface", intf, "ofport_request=%s" % port)

    def add_controller(self, controller):
        self.clist.append(
            (
                self.name + controller.name,
                "%s:%s:%d" % (controller.protocol, controller.IP(), controller.port),
            )
        )
        if self.listenPort:
            self.clist.append((self.name + "-listen", "ptcp:%s" % self.listenPort))
        ccmd = '-- --id=@%s create Controller target=\\"%s\\"'
        if self.reconnectms:
            ccmd += " max_backoff=%d" % self.reconnectms
        for param, value in self.controller_params.items():
            ccmd += " %s=%s" % (param, value)
        cargs = " ".join(ccmd % (name, target) for name, target in self.clist)
        # Controller ID list
        cids = ",".join("@%s" % name for name, _target in self.clist)
        # One ovs-vsctl command to rule them all!
        self.vsctl(cargs + " -- set bridge %s controller=[%s]" % (self, cids))

    def start(self, controllers):
        # Transcluded from Mininet source, since need to insert
        # controller parameters at switch creation time.
        int(self.dpid, 16)  # DPID must be a hex string
        switch_intfs = [
            intf for intf in self.intfList() if self.ports[intf] and not intf.IP()
        ]
        # Command to add interfaces
        intfs = " ".join(
            " -- add-port %s %s" % (self, intf) + self.intfOpts(intf)
            for intf in switch_intfs
        )
        # Command to create controller entries
        self.clist = [
            (self.name + c.name, "%s:%s:%d" % (c.protocol, c.IP(), c.port))
            for c in controllers
        ]
        if self.listenPort:
            self.clist.append((self.name + "-listen", "ptcp:%s" % self.listenPort))
        ccmd = '-- --id=@%s create Controller target=\\"%s\\"'
        if self.reconnectms:
            ccmd += " max_backoff=%d" % self.reconnectms
        for param, value in self.controller_params.items():
            ccmd += " %s=%s" % (param, value)
        cargs = " ".join(ccmd % (name, target) for name, target in self.clist)
        # Controller ID list
        cids = ",".join("@%s" % name for name, _target in self.clist)
        # Try to delete any existing bridges with the same name
        if not self.isOldOVS():
            cargs += " -- --if-exists del-br %s" % self
        # One ovs-vsctl command to rule them all!
        self.vsctl(
            cargs
            + " -- add-br %s" % self
            + " -- set bridge %s controller=[%s]" % (self, cids)
            + self.bridgeOpts()
            + intfs
        )
        # switch interfaces on mininet host, must have no IP config.
        sysctls = []
        ipcmds = []
        for intf in switch_intfs:
            ipcmds.append("addr flush dev %s" % intf)
            sysctls.append("net.ipv6.conf.%s.disable_ipv6=1" % intf)
        with tempfile.TemporaryDirectory() as tmpdir:
            sysctl_filename = os.path.join(tmpdir, "sysctl.txt")
            ip_batch_filename = os.path.join(tmpdir, "ipbatch.txt")
            with open(sysctl_filename, "w") as sysctl_file:
                sysctl_file.write("\n".join(sysctls))
            with open(ip_batch_filename, "w") as ip_batch_file:
                ip_batch_file.write("\n".join(ipcmds))
            self.cmd("sysctl -p %s" % sysctl_filename)
            result = self.cmd("ip -b %s" % ip_batch_filename)
            assert result == "", (result, ipcmds)
        # If necessary, restore TC config overwritten by OVS
        if not self.batch:
            for intf in self.intfList():
                self.TCReapply(intf)


class NoControllerFaucetSwitch(FaucetSwitch):
    """A switch without any controllers (typically for remapping hardware to software."""

    def start(self, _controllers):
        super().start(controllers=[])


class FaucetSwitchTopo(Topo):
    """FAUCET switch topology that contains a software switch."""

    CPUF = 0.5
    DELAY = "1ms"

    def __init__(self, *args, **kwargs):
        self.dpid_names = {}  # maps dpids to switch names
        self.switch_dpids = {}  # maps switch names to dpids
        self.switch_ports = {}  # maps switch names to port lists
        self.dpid_port_host = {}  # maps switch hosts to ports
        super().__init__(*args, **kwargs)

    @staticmethod
    def _get_sid_prefix(ports_served):
        """Return a unique switch/host prefix for a test."""
        # Linux tools require short interface names.
        id_chars = "".join(
            sorted(string.ascii_letters + string.digits)
        )  # pytype: disable=module-attr
        id_a = int(ports_served / len(id_chars))
        id_b = ports_served - (id_a * len(id_chars))
        return "%s%s" % (id_chars[id_a], id_chars[id_b])

    def _add_tagged_host(self, sid_prefix, tagged_vids, host_n):
        """Add a single tagged test host."""
        host_name = "t%s%1.1u" % (sid_prefix, host_n + 1)
        return self.addHost(
            name=host_name, cls=VLANHost, vlans=tagged_vids, cpu=self.CPUF
        )

    def _add_untagged_host(self, sid_prefix, host_n, in_namespace=True):
        """Add a single untagged test host."""
        host_name = "u%s%1.1u" % (sid_prefix, host_n + 1)
        return self.addHost(
            name=host_name, cls=FaucetHost, cpu=self.CPUF, inNamespace=in_namespace
        )

    def _add_extended_host(self, sid_prefix, host_n, e_cls, tmpdir):
        """Add a single extended test host."""
        host_name = "e%s%1.1u" % (sid_prefix, host_n + 1)
        return self.addHost(name=host_name, cls=e_cls, host_n=host_n, tmpdir=tmpdir)

    def _add_faucet_switch(self, sid_prefix, dpid, hw_dpid, ovs_type):
        """Add a FAUCET switch."""
        switch_cls = FaucetSwitch
        switch_name = "s%s" % sid_prefix
        self.switch_dpids[switch_name] = dpid
        self.dpid_names[dpid] = switch_name
        if hw_dpid and hw_dpid == dpid:
            remap_dpid = str(int(dpid) + 1)
            output(
                "bridging hardware switch DPID %s (%x) dataplane via OVS DPID %s (%x)\n"
                % (dpid, int(dpid), remap_dpid, int(remap_dpid))
            )
            dpid = remap_dpid
            switch_cls = NoControllerFaucetSwitch
        return self.addSwitch(
            name=switch_name,
            cls=switch_cls,
            datapath=ovs_type,
            dpid=mininet_test_util.mininet_dpid(dpid),
        )

    # Hardware switch port virtualization through
    # transparent OVS attachment bridge/patch panel
    #
    # Since FAUCET is talking to the hardware switch, it needs
    # to use the hardware switch's OpenFlow ports, rather than
    # the OpenFlow ports of the (transparent) OVS attachment bridge.

    def hw_remap_port(self, dpid, port):
        """Map OVS attachment bridge port number -> HW port number if necessary"""
        if dpid != self.hw_dpid:
            return port
        assert self.hw_ports
        return self.hw_ports[port - self.start_port]

    peer_link = namedtuple("peer_link", "port peer_dpid peer_port")

    def hw_remap_peer_link(self, dpid, link):
        """Remap HW port numbers -> OVS port numbers in link if necessary"""
        port = self.hw_remap_port(dpid, link.port)
        peer_port = self.hw_remap_port(link.peer_dpid, link.peer_port)
        return self.peer_link(port, link.peer_dpid, peer_port)

    def dpid_ports(self, dpid):
        """Return port list for dpid, remapping if necessary"""
        name = self.dpid_names[dpid]
        ports = self.switch_ports[name]
        return [self.hw_remap_port(dpid, port) for port in ports]

    @staticmethod
    def extend_port_order(port_order=None, max_length=16):
        """Extend port_order to max_length if needed"""
        if not port_order:
            port_order = []
        return port_order + list(range(len(port_order), max_length + 1))

    def _add_links(self, switch, dpid, hosts, links_per_host):
        self.switch_ports.setdefault(switch, [])
        self.dpid_port_host.setdefault(int(dpid), {})
        index = 0
        for host in hosts:
            for _ in range(links_per_host):
                # Order of switch/host is important, since host may be in a container.
                port = self.start_port + self.port_order[index]
                self.addLink(switch, host, port1=port, delay=self.DELAY, use_htb=True)
                # Keep track of switch ports
                self.switch_ports.setdefault(switch, [])
                self.switch_ports[switch].append(port)
                self.dpid_port_host[int(dpid)][port] = host
                index += 1
        return index

    # pylint: disable=too-many-locals,arguments-differ
    def build(
        self,
        ovs_type,
        ports_sock,
        test_name,
        dpids,
        n_tagged=0,
        tagged_vid=100,
        n_untagged=0,
        links_per_host=0,
        n_extended=0,
        e_cls=None,
        tmpdir=None,
        hw_dpid=None,
        switch_map=None,
        host_namespace=None,
        start_port=SWITCH_START_PORT,
        port_order=None,
        get_serialno=mininet_test_util.get_serialno,
    ):
        if not host_namespace:
            host_namespace = {}
        self.hw_dpid = hw_dpid
        self.hw_ports = sorted(switch_map) if switch_map else []
        self.start_port = start_port
        maxlength = n_tagged + n_untagged + n_extended
        self.port_order = self.extend_port_order(port_order, maxlength)
        for dpid in dpids:
            serialno = get_serialno(ports_sock, test_name)
            sid_prefix = self._get_sid_prefix(serialno)
            tagged = [
                self._add_tagged_host(sid_prefix, [tagged_vid], host_n)
                for host_n in range(n_tagged)
            ]
            untagged = [
                self._add_untagged_host(
                    sid_prefix, host_n, host_namespace.get(host_n, True)
                )
                for host_n in range(n_untagged)
            ]
            extended = [
                self._add_extended_host(sid_prefix, host_n, e_cls, tmpdir)
                for host_n in range(n_extended)
            ]
            switch = self._add_faucet_switch(sid_prefix, dpid, hw_dpid, ovs_type)
            self._add_links(switch, dpid, tagged + untagged + extended, links_per_host)


class BaseFAUCET(Controller):
    """Base class for FAUCET and Gauge controllers."""

    # Set to True to have cProfile output to controller log.
    CPROFILE = False
    controller_intf = None
    controller_ipv6 = False
    controller_ip = None
    pid_file = None
    tmpdir = None
    ofcap = None
    MAX_OF_PKTS = 5000
    MAX_CTL_TIME = 600

    BASE_CARGS = " ".join(("--verbose", "--use-stderr", "--ryu-ofp-tcp-listen-port=%s"))

    RYU_CONF = """
[DEFAULT]
echo_request_interval=10
maximum_unreplied_echo_requests=5
socket_timeout=15
"""

    def __init__(
        self,
        name,
        tmpdir,
        controller_intf=None,
        controller_ipv6=False,
        cargs="",
        **kwargs,
    ):
        self.name_no_pid = name
        name_with_pid = "%s-%u" % (name, os.getpid())
        self.tmpdir = tmpdir
        self.controller_intf = controller_intf
        self.controller_ipv6 = controller_ipv6
        self.cached_ryu_pid = None
        super().__init__(
            name_with_pid, cargs=self._add_cargs(cargs, name_with_pid), **kwargs
        )

    def _add_cargs(self, cargs, name):
        ofp_listen_host_arg = ""
        if self.controller_intf is not None:
            socket_type = socket.AF_INET
            if self.controller_ipv6:
                socket_type = socket.AF_INET6
            self.controller_ip = (
                netifaces.ifaddresses(  # pylint: disable=c-extension-no-member
                    self.controller_intf
                )[socket_type][0]["addr"]
            )
            ofp_listen_host_arg = "--ryu-ofp-listen-host=%s" % self.controller_ip
        self.pid_file = os.path.join(self.tmpdir, name + ".pid")
        pid_file_arg = "--ryu-pid-file=%s" % self.pid_file
        ryu_conf_file = os.path.join(self.tmpdir, "ryu.conf")
        with open(ryu_conf_file, "w", encoding="utf-8") as ryu_conf:
            ryu_conf.write(self.RYU_CONF)
        ryu_conf_arg = "--ryu-config-file=%s" % ryu_conf_file
        return " ".join(
            (self.BASE_CARGS, pid_file_arg, ryu_conf_arg, ofp_listen_host_arg, cargs)
        )

    def IP(self):  # pylint: disable=invalid-name,arguments-differ
        if self.controller_intf is not None:
            return self.controller_ip
        return super().IP()

    def _start_tcpdump(self):
        """Start a tcpdump for OF port."""
        self.ofcap = os.path.join(
            self.tmpdir, "-".join((self.name, "%u-of.cap" % self.port))
        )
        tcpdump_args = " ".join(
            (
                "-s 0",
                "-e",
                "-n",
                "-U",
                "-q",
                "-W 1",  # max files 1
                "-G %u" % (self.MAX_CTL_TIME - 1),
                "-c %u" % (self.MAX_OF_PKTS),
                "-i %s" % self.controller_intf,
                "-w %s" % self.ofcap,
                "tcp and port %u" % self.port,
                ">/dev/null",
                "2>/dev/null",
            )
        )
        self.cmd("timeout %s tcpdump %s &" % (self.MAX_CTL_TIME, tcpdump_args))
        for _ in range(5):
            if os.path.exists(self.ofcap):
                return
            time.sleep(1)
        assert False, "tcpdump of OF channel did not start"

    @staticmethod
    def _tls_cargs(ofctl_port, ctl_privkey, ctl_cert, ca_certs):
        """Add TLS/cert parameters to Ryu."""
        tls_cargs = []
        for carg_val, carg_key in (
            (ctl_privkey, "ryu-ctl-privkey"),
            (ctl_cert, "ryu-ctl-cert"),
            (ca_certs, "ryu-ca-certs"),
        ):
            if carg_val:
                tls_cargs.append(("--%s=%s" % (carg_key, carg_val)))
        if tls_cargs:
            tls_cargs.append(("--ryu-ofp-ssl-listen-port=%u" % ofctl_port))
        return " ".join(tls_cargs)

    def _command(self, env, name, args):
        """Wrap controller startup command in shell script with environment."""
        env_vars = []
        for var, val in sorted(env.items()):
            env_vars.append("=".join((var, val)))
        cprofile_args = ""
        if self.CPROFILE:
            cprofile_args = "python3 -m cProfile -s time"
        full_faucet_dir = os.path.abspath(mininet_test_util.FAUCET_DIR)
        faucet_cli = "PYTHONPATH=%s %s timeout %u %s %s %s" % (
            os.path.dirname(full_faucet_dir),
            " ".join(env_vars),
            self.MAX_CTL_TIME,
            os.path.join(full_faucet_dir, "__main__.py"),
            cprofile_args,
            args,
        )
        return faucet_cli

    def ryu_pid(self):
        """Return PID of ryu-manager process."""
        if self.cached_ryu_pid is None:
            if os.path.exists(self.pid_file) and os.path.getsize(self.pid_file) > 0:
                with open(self.pid_file, encoding="utf-8") as pid_file:
                    self.cached_ryu_pid = int(pid_file.read())
        return self.cached_ryu_pid

    def listen_port(self, port, state="LISTEN"):
        """Return True if port in specified TCP state."""
        ryu_pid = self.ryu_pid()
        if ryu_pid is not None:
            for ipv in (4, 6):
                listening_pids = [
                    int(pid)
                    for pid in self.cmd(
                        mininet_test_util.tcp_listening_cmd(
                            port, ipv=ipv, state=state, pid=ryu_pid
                        )
                    ).split()
                ]
                if listening_pids:
                    return True
        return False

    # pylint: disable=invalid-name
    @staticmethod
    def checkListening():
        """Mininet's checkListening() causes occasional false positives (with
        exceptions we can't catch), and we handle port conflicts ourselves anyway."""
        return

    def listening(self):
        """Return True if controller listening on required ports."""
        return self.listen_port(self.port)

    def connected(self):
        """Return True if at least one switch connected and controller healthy."""
        return self.listen_port(self.port, state="ESTABLISHED") and self.healthy()

    def logname(self):
        """Return log file for controller."""
        return os.path.join("/tmp", self.name + ".log")

    def healthy(self):
        """Return True if controller logging and listening on required ports."""
        if (
            os.path.exists(self.logname())
            and os.path.getsize(self.logname())
            and self.listening()
        ):
            return True
        return False

    def start(self):
        """Start tcpdump for OF port and then start controller."""
        self._start_tcpdump()
        super().start()

    def _stop_cap(self):
        """Stop tcpdump for OF port."""
        if os.path.exists(self.ofcap):
            self.cmd(" ".join(["fuser", "-15", "-k", self.ofcap]))

    def stop(self):  # pylint: disable=arguments-differ
        """Stop controller."""
        try:
            if self.CPROFILE:
                os.kill(self.ryu_pid(), 2)
            else:
                os.kill(self.ryu_pid(), 15)
        except ProcessLookupError:
            pass
        self._stop_cap()
        super().stop()
        if os.path.exists(self.logname()):
            tmpdir_logname = os.path.join(self.tmpdir, self.name + "-stdout-stderr.log")
            if os.path.exists(tmpdir_logname):
                os.remove(tmpdir_logname)
            shutil.move(self.logname(), tmpdir_logname)


class FAUCET(BaseFAUCET):
    """Start a FAUCET controller."""

    START_ARGS = [
        "--ryu-app-lists=%s"
        % (os.path.dirname(os.path.realpath(__file__)) + "/../ofctl_rest/ofctl_rest.py")
    ]

    def __init__(
        self,
        name,
        tmpdir,
        controller_intf,
        controller_ipv6,
        env,
        ctl_privkey,
        ctl_cert,
        ca_certs,
        ports_sock,
        prom_port,
        port,
        test_name,
        **kwargs,
    ):
        self.prom_port = prom_port
        self.ofctl_port = mininet_test_util.find_free_port(ports_sock, test_name)
        env["OFCTL_PORT"] = str(self.ofctl_port)
        env["OFCTL_HOST"] = mininet_test_util.LOCALHOSTV6
        cargs = " ".join(self._tls_cargs(port, ctl_privkey, ctl_cert, ca_certs))
        super().__init__(
            name,
            tmpdir,
            controller_intf,
            controller_ipv6,
            cargs=cargs,
            command=self._command(env, name, " ".join(self.START_ARGS)),
            port=port,
            **kwargs,
        )

    def listening(self):
        return (
            self.listen_port(self.ofctl_port)
            and self.listen_port(self.prom_port)
            and super().listening()
        )


class Gauge(BaseFAUCET):
    """Start a Gauge controller."""

    def __init__(
        self,
        name,
        tmpdir,
        controller_intf,
        controller_ipv6,
        env,
        ctl_privkey,
        ctl_cert,
        ca_certs,
        port,
        **kwargs,
    ):
        super().__init__(
            name,
            tmpdir,
            controller_intf,
            controller_ipv6,
            cargs=self._tls_cargs(port, ctl_privkey, ctl_cert, ca_certs),
            command=self._command(env, name, "--gauge"),
            port=port,
            **kwargs,
        )



"""Mininet tests for clib."""


import os
import re

from clib import mininet_test_base

from clib.tcpdump_helper import TcpdumpHelper
from clib.docker_host import make_docker_host


class FaucetSimpleTest(mininet_test_base.FaucetTestBase):
    """Basic untagged VLAN test."""

    N_UNTAGGED = 4
    N_TAGGED = 0
    LINKS_PER_HOST = 1
    CONFIG_GLOBAL = """
vlans:
    100:
        description: "untagged"
"""

    CONFIG = """
        interfaces:
            %(port_1)d:
                native_vlan: 100
                description: "b1"
            %(port_2)d:
                native_vlan: 100
                description: "b2"
            %(port_3)d:
                native_vlan: 100
                description: "b3"
            %(port_4)d:
                native_vlan: 100
                description: "b4"
"""

    def setUp(self):
        super().setUp()
        self.topo = self.topo_class(
            self.OVS_TYPE,
            self.ports_sock,
            self._test_name(),
            [self.dpid],
            n_tagged=self.N_TAGGED,
            n_untagged=self.N_UNTAGGED,
            n_extended=self.N_EXTENDED,
            e_cls=self.EXTENDED_CLS,
            tmpdir=self.tmpdir,
            links_per_host=self.LINKS_PER_HOST,
            hw_dpid=self.hw_dpid,
        )
        self.start_net()

    def test_ping_all(self):
        """All hosts should have connectivity."""
        self.ping_all_when_learned()


class FaucetTcpdumpHelperTest(FaucetSimpleTest):
    """Test for TcpdumpHelper class"""

    def _terminate_with_zero(self, tcpdump_helper):
        term_returns = tcpdump_helper.terminate()
        self.assertEqual(0, term_returns, msg="terminate code not 0: %d" % term_returns)

    def _terminate_with_nonzero(self, tcpdump_helper):
        term_returns = tcpdump_helper.terminate()
        self.assertNotEqual(
            0, term_returns, msg="terminate code is 0: %d" % term_returns
        )

    def test_tcpdump_execute(self):
        """Check tcpdump filter monitors ping using execute"""
        self.ping_all_when_learned()
        from_host = self.net.hosts[0]
        to_host = self.net.hosts[1]
        tcpdump_filter = "icmp"
        tcpdump_helper = TcpdumpHelper(
            to_host,
            tcpdump_filter,
            [lambda: from_host.cmd("ping -c1 %s" % to_host.IP())],
        )
        tcpdump_txt = tcpdump_helper.execute()
        self.assertTrue(re.search("%s: ICMP echo request" % to_host.IP(), tcpdump_txt))
        self._terminate_with_zero(tcpdump_helper)

    def test_tcpdump_pcap(self):
        """Check tcpdump creates pcap output"""
        self.ping_all_when_learned()
        from_host = self.net.hosts[0]
        to_host = self.net.hosts[1]
        tcpdump_filter = "icmp"
        pcap_file = os.path.join(self.tmpdir, "out.pcap")
        tcpdump_helper = TcpdumpHelper(
            to_host,
            tcpdump_filter,
            [lambda: from_host.cmd("ping -c3 %s" % to_host.IP())],
            pcap_out=pcap_file,
            packets=None,
        )
        tcpdump_helper.execute()
        self._terminate_with_zero(tcpdump_helper)
        result = from_host.cmd("tcpdump -en -r %s" % pcap_file)
        self.assertEqual(result.count("ICMP echo reply"), 3, "three icmp echo replies")

    def test_tcpdump_noblock(self):
        """Check tcpdump uses nonblocking next_line"""
        self.ping_all_when_learned()
        from_host = self.net.hosts[0]
        to_host = self.net.hosts[1]
        tcpdump_filter = "icmp"
        tcpdump_helper = TcpdumpHelper(
            to_host,
            tcpdump_filter,
            [lambda: from_host.cmd("ping -c10 %s" % to_host.IP())],
            blocking=False,
            packets=None,
        )
        count = 0
        while tcpdump_helper.next_line():
            count = count + 1
            self.assertTrue(count < 10, "Too many ping results before noblock")
        self._terminate_with_nonzero(tcpdump_helper)

    def test_tcpdump_nextline(self):
        """Check tcpdump filter monitors ping using next_line"""
        self.ping_all_when_learned()
        from_host = self.net.hosts[0]
        to_host = self.net.hosts[1]
        tcpdump_filter = "icmp"
        tcpdump_helper = TcpdumpHelper(
            to_host,
            tcpdump_filter,
            [lambda: from_host.cmd("ping -c5 -i2 %s" % to_host.IP())],
        )

        self.assertTrue(re.search("proto ICMP", tcpdump_helper.next_line()))
        next_line = tcpdump_helper.next_line()
        self.assertTrue(
            re.search("%s: ICMP echo request" % to_host.IP(), next_line), next_line
        )
        self.assertTrue(re.search("proto ICMP", tcpdump_helper.next_line()))
        next_line = tcpdump_helper.next_line()
        self.assertTrue(
            re.search("%s: ICMP echo reply" % from_host.IP(), next_line), next_line
        )
        self.assertFalse(re.search("ICMP", tcpdump_helper.next_line()))
        while tcpdump_helper.next_line():
            pass
        self._terminate_with_zero(tcpdump_helper)


class FaucetDockerHostTest(FaucetSimpleTest):
    """Test basic docker host functionality"""

    N_UNTAGGED = 2
    N_EXTENDED = 2
    EXTENDED_CLS = make_docker_host("faucet/test-host", startup_timeout_ms=90 * 1e3)

    def test_containers(self):
        """Test containers to make sure they're actually docker."""
        count = 0
        host_name = None

        for host in self.net.hosts:
            marker = host.cmd("cat /root/test_marker.txt").strip()
            if marker == "faucet-test-host":
                host_name = host.name
                count = count + 1
                host.activate()
                host.wait()

        self.assertTrue(
            count == self.N_EXTENDED,
            "Found %d containers, expected %d" % (count, self.N_EXTENDED),
        )

        self.assertTrue(
            os.path.exists(os.path.join(self.tmpdir, host_name, "tmp")),
            "container tmp dir missing",
        )

        host_log = os.path.join(self.tmpdir, host_name, "activate.log")
        with open(host_log, "r", encoding="utf-8") as host_log_file:
            lines = host_log_file.readlines()
            output = " ".join(lines).strip()
            self.assertEqual(output, "hello faucet")



#!/usr/bin/env python3

"""Report state based on FAUCET/Gauge/Prometheus variables."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pytype: disable=pyi-error
# pytype: disable=import-error
import argparse
import sys
import time
import urllib.error
import urllib.parse
import urllib.request
import requests
from prometheus_client import parser


# TODO: byte/packet counters could be per second (given multiple samples)
def decode_value(metric_name, value):
    """Convert values to human readible format based on metric name"""
    result = value
    if metric_name == "learned_macs":
        result = ":".join(
            format(octet, "02x")
            for octet in int(value).to_bytes(  # pytype: disable=attribute-error
                6, byteorder="big"
            )
        )
    return result


def scrape_prometheus(endpoints, retries=3, err_output_file=sys.stdout):
    """Scrape a list of Prometheus/FAUCET/Gauge endpoints and aggregate results."""
    metrics = []
    for endpoint in endpoints:
        content = None
        err = None
        for _ in range(retries):
            try:
                if endpoint.startswith("http"):
                    response = requests.get(endpoint, timeout=30)
                    if (
                        response.status_code
                        == requests.status_codes.codes.ok  # pylint: disable=no-member
                    ):
                        content = response.content.decode("utf-8", "strict")
                        break
                else:
                    with urllib.request.urlopen(  # pytype: disable=module-attr
                        endpoint
                    ) as response:
                        content = response.read().decode("utf-8", "strict")
                    break
            except (requests.exceptions.ConnectionError, ValueError) as exception:
                err = exception
                time.sleep(1)
        if err is not None:
            err_output_file.write(str(err))
            return None
        try:
            endpoint_metrics = parser.text_string_to_metric_families(content)
            metrics.extend(endpoint_metrics)
        except ValueError as err:
            err_output_file.write(str(err))
            return None
    return metrics


def _get_samples_from_metrics(metrics, metric_name, label_matches, nonzero_only=False):
    result = []
    for metric in metrics:
        if metric_name is None or metric.name == metric_name:
            for sample in metric.samples:
                labels = sample.labels
                value = sample.value
                if label_matches is None or set(label_matches.items()).issubset(
                    set(labels.items())
                ):
                    if nonzero_only and int(value) == 0:
                        continue
                    result.append(sample)
    return result


def get_samples(endpoints, metric_name, label_matches, nonzero_only=False, retries=3):
    """return a list of Prometheus samples for a given metric

    Prometheus Sample objects are named tuples with the fields: name, labels,
    value, timestamp, exemplar.

    Arguments:
        endpoints (list of strings): the prometheus endpoints to query
        metric_name (string): the metric to retrieve
        label_matches (dict): filters results by label
        nonzero_only (bool): only return samples with non-zero values
        retries (int): number of retries when querying
    Returns:
        list of Prometheus Sample objects"""
    metrics = scrape_prometheus(endpoints, retries)
    if metrics is None:
        return None
    return _get_samples_from_metrics(metrics, metric_name, label_matches, nonzero_only)


def report_label_match_metrics(
    report_metrics,
    metrics,
    display_labels=None,
    nonzero_only=False,
    delim="\t",
    label_matches=None,
):
    """Text report on a list of Prometheus metrics."""
    report_output = []
    if report_metrics is None:
        report_metrics = [None]
    for metric_name in report_metrics:
        for sample in _get_samples_from_metrics(
            metrics, metric_name, label_matches, nonzero_only
        ):
            labels = sample.labels
            value = sample.value
            sorted_labels = [
                (key, val)
                for key, val in sorted(labels.items())
                if not display_labels or key in display_labels
            ]
            value = decode_value(metric_name, value)
            report_output.append(
                delim.join((metric_name, str(sorted_labels), str(value)))
            )
        report_output = "\n".join(report_output)
    return report_output


def parse_args(sys_args):
    """Parse and return CLI args."""

    arg_parser = argparse.ArgumentParser(
        prog="fctl",
        description="Retrieve FAUCET/Gauge state using Prometheus variables.",
        usage="""
    MACs learned on a DP.

    {argv0} -n --endpoints=http://172.17.0.1:9302 --metrics=learned_macs --labels=dp_id:0xb827eb608918

    Status of all DPs

    {argv0} -n --endpoints=http://172.17.0.1:9302 --metrics=dp_status
""".format(
            **{"argv0": sys.argv[0]}
        ),
    )
    arg_parser.add_argument(
        "-n", "--nonzero", action="store_true", help="nonzero results only"
    )
    arg_parser.add_argument("-e", "--endpoints", help="list of endpoint URLs to query")
    arg_parser.add_argument("-m", "--metrics", help="list of metrics to query")
    arg_parser.add_argument(
        "-l", "--labels", help="list of labels that must be present"
    )
    arg_parser.add_argument(
        "--display-labels", help="list of labels to filter display by (default all)"
    )

    endpoints = []
    report_metrics = []
    label_matches = None
    display_labels = None
    nonzero_only = False

    try:
        args = arg_parser.parse_args(sys_args)
        if args.nonzero:
            nonzero_only = True
        if args.endpoints:
            endpoints = args.endpoints.split(",")
        if args.metrics:
            report_metrics = args.metrics.split(",")
        if args.labels:
            label_matches = {}
            for label_value in args.labels.split(","):
                label, value = label_value.split(":")
                label_matches[label] = value
        if args.display_labels:
            display_labels = args.display_labels.split(",")
    except (KeyError, IndexError):
        arg_parser.print_usage()
        sys.exit(-1)

    return (endpoints, report_metrics, label_matches, nonzero_only, display_labels)


def main():
    (
        endpoints,
        report_metrics,
        label_matches,
        nonzero_only,
        display_labels,
    ) = parse_args(sys.argv[1:])
    metrics = scrape_prometheus(endpoints)
    if metrics is None:
        sys.exit(1)
    report = report_label_match_metrics(
        report_metrics,
        metrics,
        nonzero_only=nonzero_only,
        label_matches=label_matches,
        display_labels=display_labels,
    )
    print(report)


if __name__ == "__main__":
    main()



"""Manages movement of packets through the faucet pipeline."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import functools
import faucet.faucet_metadata as faucet_md
from faucet import valve_of
from faucet.valve_manager_base import ValveManagerBase


class ValvePipeline(ValveManagerBase):
    """Responsible for maintaing the integrity of the Faucet pipeline for a
    single valve.

    Controls what packets a module sees in its tables and how it can pass
    packets through the pipeline.

    Responsible for installing flows in the vlan, egress and classification
    tables"""

    def __init__(self, dp):
        self.dp = dp
        self.vlan_table = dp.tables["vlan"]
        self.classification_table = dp.classification_table()
        self.output_table = dp.output_table()
        self.egress_table = None
        self.egress_acl_table = None
        if dp.egress_pipeline:
            self.egress_table = dp.tables["egress"]
            self.egress_acl_table = dp.tables.get("egress_acl")
        self.filter_priority = self._FILTER_PRIORITY
        self.select_priority = self._HIGH_PRIORITY

    @staticmethod
    @functools.lru_cache(maxsize=1024)
    def _accept_to_table(table, actions):
        inst = [table.goto_this()]
        if actions:
            inst.append(valve_of.apply_actions(actions))
        return tuple(inst)

    @functools.lru_cache(maxsize=1024)
    def accept_to_vlan(self, actions=None):
        """Get instructions to forward packet through the pipeline to
        vlan table.
        args:
            actions: (optional) list of actions to apply to packet.
        returns:
            list of instructions
        """
        return self._accept_to_table(self.vlan_table, actions)

    @functools.lru_cache(maxsize=1024)
    def accept_to_classification(self, actions=None):
        """Get instructions to forward packet through the pipeline to
        classification table.
        args:
            actions: (optional) list of actions to apply to packet.
        returns:
            list of instructions
        """
        return self._accept_to_table(self.classification_table, actions)

    @functools.lru_cache(maxsize=1024)
    def accept_to_l2_forwarding(self, actions=None):
        """Get instructions to forward packet through the pipeline to l2
        forwarding.
        args:
            actions: (optional) list of actions to apply to packet.
        returns:
            list of instructions
        """
        return self._accept_to_table(self.output_table, actions)

    @functools.lru_cache(maxsize=1024)
    def accept_to_egress(self, actions=None):
        """Get instructions to forward packet through the pipeline to egress
        table

        Raises an assertion error if egress pipeline is not configured

        args:
            actions: (optional) list of actions to apply to the packet
        returns:
            list of instructions:
        """
        assert self.egress_table is not None
        return self._accept_to_table(self.egress_table, actions)

    def output(self, port, vlan, hairpin=False, external_forwarding_requested=None):
        """Get instructions list to output a packet through the regular
        pipeline.

        args:
            port: Port object of port to output packet to
            vlan: Vlan object of vlan to output packet on
            hairpin: if True, hairpinning is required
            apply_egress_acl: if True the packet will be sent to the egress acl
                table before being output
        returns:
            list of Instructions
        """
        instructions = []
        if self.egress_table:
            metadata, metadata_mask = faucet_md.get_egress_metadata(
                port.number, vlan.vid
            )
            if self.egress_acl_table:
                instructions.extend(
                    valve_of.metadata_goto_table(
                        metadata, metadata_mask, self.egress_acl_table
                    )
                )
            else:
                instructions.extend(
                    valve_of.metadata_goto_table(
                        metadata, metadata_mask, self.egress_table
                    )
                )
        else:
            instructions.append(
                valve_of.apply_actions(
                    vlan.output_port(
                        port,
                        hairpin=hairpin,
                        output_table=self.output_table,
                        external_forwarding_requested=external_forwarding_requested,
                    )
                )
            )
        return tuple(instructions)

    def initialise_tables(self):
        """Install rules to initialise the classification_table"""
        ofmsgs = []
        # drop broadcast sources
        if self.dp.drop_broadcast_source_address:
            ofmsgs.extend(self.filter_packets({"eth_src": valve_of.mac.BROADCAST_STR}))

        ofmsgs.extend(
            self.filter_packets(
                {"eth_type": valve_of.ECTP_ETH_TYPE}, priority_offset=10
            )
        )

        return ofmsgs

    def _add_egress_table_rule(self, port, vlan, pop_vlan=True):
        metadata, metadata_mask = faucet_md.get_egress_metadata(port.number, vlan.vid)
        actions = copy.copy(port.mirror_actions())
        if pop_vlan:
            actions.append(valve_of.pop_vlan())
        actions.append(valve_of.output_port(port.number))
        inst = (valve_of.apply_actions(tuple(actions)),)
        return self.egress_table.flowmod(
            self.egress_table.match(
                vlan=vlan, metadata=metadata, metadata_mask=metadata_mask
            ),
            priority=self.dp.high_priority,
            inst=inst,
        )

    def add_port(self, port):
        ofmsgs = []
        if self.egress_table is None:
            return ofmsgs
        for vlan in port.tagged_vlans:
            ofmsgs.append(self._add_egress_table_rule(port, vlan, pop_vlan=False))
        if port.native_vlan is not None:
            ofmsgs.append(self._add_egress_table_rule(port, port.native_vlan))
        return ofmsgs

    def del_port(self, port):
        ofmsgs = []
        if self.egress_table:
            mask = faucet_md.PORT_METADATA_MASK
            ofmsgs.append(
                self.egress_table.flowdel(
                    self.egress_table.match(
                        metadata=port.number & mask, metadata_mask=mask
                    )
                )
            )
        return ofmsgs

    def filter_packets(self, match_dict, priority_offset=0):
        """get a list of flow modification messages to filter packets from
        the pipeline.
        args:
            match_dict: a dictionary specifying the match fields
            priority_offset: used to prevent overlapping entries
        """
        return [
            self.classification_table.flowdrop(
                self.classification_table.match(**match_dict),
                priority=self.filter_priority + priority_offset,
            )
        ]

    def select_packets(self, target_table, match_dict, actions=None, priority_offset=0):
        """retrieve rules to redirect packets matching match_dict to table"""
        inst = [target_table.goto_this()]
        if actions is not None:
            inst.append(valve_of.apply_actions(actions))
        return [
            self.classification_table.flowmod(
                self.classification_table.match(**match_dict),
                priority=self.select_priority + priority_offset,
                inst=tuple(inst),
            )
        ]

    def remove_filter(self, match_dict, strict=True, priority_offset=0):
        """retrieve flow mods to remove a filter from the classification table"""
        priority = None
        if strict:
            priority = self.filter_priority + priority_offset
        return [
            self.classification_table.flowdel(
                self.classification_table.match(**match_dict),
                priority=priority,
                strict=strict,
            )
        ]



"""Deprecated OF matches."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Map of old field name, to new.
OLD_MATCH_FIELDS = {
    "dl_dst": "eth_dst",
    "dl_src": "eth_src",
    "dl_type": "eth_type",
    "dl_vlan": "vlan_vid",
    "nw_proto": "ip_proto",
    "nw_src": "ipv4_src",
    "nw_dst": "ipv4_dst",
}



#!/usr/bin/env python3

"""Launch forwarder script for Faucet/Gauge"""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import sys

from pbr.version import VersionInfo

if sys.version_info < (3,) or sys.version_info < (3, 5):
    raise ImportError(
        """You are trying to run faucet on python {py}

Faucet is not compatible with python {py}, please upgrade to python 3.5 or newer.""".format(
            py=".".join([str(v) for v in sys.version_info[:3]])
        )
    )

RYU_OPTIONAL_ARGS = [
    ("ca-certs", "CA certificates"),
    (
        "config-dir",
        """Path to a config directory to pull `*.conf` files
                      from. This file set is sorted, so as to provide a
                      predictable parse order if individual options are
                      over-ridden. The set is parsed after the file(s)
                      specified via previous --config-file, arguments hence
                      over-ridden options in the directory take precedence.""",
    ),
    (
        "config-file",
        """Path to a config file to use. Multiple config files
                       can be specified, with values in later files taking
                       precedence. Defaults to None.""",
        "/etc/faucet/ryu.conf",
    ),
    ("ctl-cert", "controller certificate"),
    ("ctl-privkey", "controller private key"),
    ("default-log-level", "default log level"),
    ("log-config-file", "Path to a logging config file to use"),
    ("log-dir", "log file directory"),
    ("log-file", "log file name"),
    ("log-file-mode", "default log file permission"),
    ("observe-links", "observe link discovery events"),
    ("ofp-listen-host", "openflow listen host (default 0.0.0.0)"),
    ("ofp-ssl-listen-port", "openflow ssl listen port (default: 6653)"),
    (
        "ofp-switch-address-list",
        """list of IP address and port pairs (default empty).
                                   e.g., "127.0.0.1:6653,[::1]:6653""",
    ),
    (
        "ofp-switch-connect-interval",
        "interval in seconds to connect to switches (default 1)",
    ),
    ("ofp-tcp-listen-port", "openflow tcp listen port (default: 6653)"),
    ("pid-file", "pid file name"),
    ("user-flags", "Additional flags file for user applications"),
]


def parse_args(sys_args):
    """Parse Faucet/Gauge arguments.

    Returns:
        argparse.Namespace: command line arguments
    """

    args = argparse.ArgumentParser(prog="faucet", description="Faucet SDN Controller")
    args.add_argument("--gauge", action="store_true", help="run Gauge instead")
    args.add_argument(
        "-v", "--verbose", action="store_true", help="produce verbose output"
    )
    args.add_argument(
        "-V", "--version", action="store_true", help="print version and exit"
    )
    args.add_argument("--use-stderr", action="store_true", help="log to standard error")
    args.add_argument("--use-syslog", action="store_true", help="output to syslog")
    args.add_argument(
        "--ryu-app-lists",
        action="append",
        help="add Ryu app (can be specified multiple times)",
        metavar="APP",
    )

    for ryu_arg in RYU_OPTIONAL_ARGS:
        if len(ryu_arg) >= 3:
            args.add_argument(
                "--ryu-%s" % ryu_arg[0], help=ryu_arg[1], default=ryu_arg[2]
            )
        else:
            args.add_argument("--ryu-%s" % ryu_arg[0], help=ryu_arg[1])

    return args.parse_args(sys_args)


def print_version():
    """Print version number and exit."""
    version = VersionInfo("c65faucet").semantic_version().release_string()
    message = "c65faucet %s" % version
    print(message)


def build_ryu_args(argv):
    args = parse_args(argv[1:])

    # Checking version number?
    if args.version:
        print_version()
        return []

    prog = os.path.basename(argv[0])
    ryu_args = []

    # Handle log location
    if args.use_stderr:
        ryu_args.append("--use-stderr")
    if args.use_syslog:
        ryu_args.append("--use-syslog")

    # Verbose output?
    if args.verbose:
        ryu_args.append("--verbose")

    for arg, val in vars(args).items():
        if not val or not arg.startswith("ryu"):
            continue
        if arg == "ryu_app_lists":
            continue
        if arg == "ryu_config_file" and not os.path.isfile(val):
            continue
        arg_name = arg.replace("ryu_", "").replace("_", "-")
        ryu_args.append("--%s=%s" % (arg_name, val))

    # Running Faucet or Gauge?
    if args.gauge or os.path.basename(prog) == "gauge":
        ryu_args.append("faucet.gauge")
    else:
        ryu_args.append("faucet.faucet")

    # Check for additional Ryu apps.
    if args.ryu_app_lists:
        ryu_args.extend(args.ryu_app_lists)

    # Replace current process with ryu-manager from PATH (no PID change).
    ryu_args.insert(0, "osken-manager")
    return ryu_args


def main():
    """Main program."""
    ryu_args = build_ryu_args(sys.argv)
    if ryu_args:
        os.execvp(ryu_args[0], ryu_args)


if __name__ == "__main__":
    main()



"""Manage flooding/learning on datapaths."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from faucet.valve_switch_standalone import (
    ValveSwitchManager,
    ValveSwitchFlowRemovedManager,
)
from faucet.valve_switch_stack import (
    ValveSwitchStackManagerNoReflection,
    ValveSwitchStackManagerReflection,
)


def valve_switch_factory(logger, dp, pipeline, stack_manager):
    """Return switch flood/learning manager based on datapath configuration.

    Args:
        logger: logger instance.
        dp: DP instance.
        pipeline: ValvePipeline instance.
    Returns:
        switch manager instance.
    """
    restricted_bcast_arpnd = bool(dp.restricted_bcast_arpnd_ports())
    eth_dst_hairpin_table = dp.tables.get("eth_dst_hairpin", None)
    vlan_acl_table = dp.tables.get("vlan_acl", None)

    switch_args = {
        "logger": logger,
        "ports": dp.ports,
        "vlans": dp.vlans,
        "vlan_table": dp.tables["vlan"],
        "vlan_acl_table": vlan_acl_table,
        "eth_src_table": dp.tables["eth_src"],
        "eth_dst_table": dp.tables["eth_dst"],
        "eth_dst_hairpin_table": eth_dst_hairpin_table,
        "flood_table": dp.tables["flood"],
        "classification_table": dp.classification_table,
        "pipeline": pipeline,
        "use_group_table": dp.group_table,
        "groups": dp.groups,
        "combinatorial_port_flood": dp.combinatorial_port_flood,
        "canonical_port_order": dp.canonical_port_order,
        "restricted_bcast_arpnd": restricted_bcast_arpnd,
        "has_externals": dp.has_externals,
        "learn_ban_timeout": dp.learn_ban_timeout,
        "learn_timeout": dp.timeout,
        "learn_jitter": dp.learn_jitter,
        "cache_update_guard_time": dp.cache_update_guard_time,
        "idle_dst": dp.idle_dst,
        "dp_high_priority": dp.high_priority,
        "dp_highest_priority": dp.highest_priority,
        "faucet_dp_mac": dp.faucet_dp_mac,
        "drop_spoofed_faucet_mac": dp.drop_spoofed_faucet_mac,
    }

    if dp.stack:
        switch_class = ValveSwitchStackManagerNoReflection
        if dp.stack.root_flood_reflection:
            switch_class = ValveSwitchStackManagerReflection
            logger.info("Using stacking root flood reflection")
        else:
            logger.info("Not using stacking root flood reflection")
        switch_args.update(
            {
                "stack_manager": stack_manager,
            }
        )
        return switch_class(**switch_args)

    switch_class = ValveSwitchManager
    if dp.use_idle_timeout:
        switch_class = ValveSwitchFlowRemovedManager
    return switch_class(**switch_args)  # pytype: disable=wrong-keyword-args



"""Utility functions to parse/create OpenFlow messages."""

# pylint: disable=too-many-lines

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import ipaddress
import random

# pylint: disable=unused-import
from os_ken.lib import mac  # noqa: F401
from os_ken.lib import ofctl_v1_3 as ofctl
from os_ken.lib.ofctl_utils import (
    str_to_int,
    to_match_ip,
    to_match_masked_int,
    to_match_eth,
    to_match_vid,
    OFCtlUtil,
)
from os_ken.ofproto import ether

# pylint: disable=unused-import
from os_ken.ofproto import inet  # noqa: F401
from os_ken.ofproto import ofproto_v1_3 as ofp
from os_ken.ofproto import ofproto_v1_3_parser as parser

from faucet.conf import test_config_condition, InvalidConfigError
from faucet.valve_of_old import OLD_MATCH_FIELDS

MIN_VID = 1
MAX_VID = 4095
VLAN_GROUP_OFFSET = MAX_VID + 1
ROUTE_GROUP_OFFSET = VLAN_GROUP_OFFSET * 2
OFP_VERSIONS = [ofp.OFP_VERSION]
OFP_IN_PORT = ofp.OFPP_IN_PORT
MAX_PACKET_IN_BYTES = 194  # largest packet is icmp6 echo req with 128 byte payload
ECTP_ETH_TYPE = 0x9000

# https://en.wikipedia.org/wiki/IEEE_P802.1p
# Avoid use of PCP 1 which is BK priority (lowest)

TUNNEL_INDICATOR_FIELD = "vlan_pcp"
# Used to indicate traffic in a one-to-one bi-directional tunnel is heading
# in the reverse/return direction
PCP_TUNNEL_REVERSE_DIRECTION_FLAG = 4
# Used to indicate traffic belongs in a tunnel (for all cases not including
# the 1-1 reverse bi-directional tunnel)
PCP_TUNNEL_FLAG = 3

PCP_EXT_PORT_FLAG = 2
PCP_NONEXT_PORT_FLAG = 0
EXTERNAL_FORWARDING_FIELD = "vlan_pcp"


OFERROR_TYPE_CODE = {
    ofp.OFPET_HELLO_FAILED: (
        "OFPET_HELLO_FAILED",
        {
            ofp.OFPHFC_INCOMPATIBLE: "OFPHFC_INCOMPATIBLE",
            ofp.OFPHFC_EPERM: "OFPHFC_EPERM",
        },
    ),
    ofp.OFPET_BAD_REQUEST: (
        "OFPET_BAD_REQUEST",
        {
            ofp.OFPBRC_BAD_VERSION: "OFPBRC_BAD_VERSION",
            ofp.OFPBRC_BAD_TYPE: "OFPBRC_BAD_TYPE",
            ofp.OFPBRC_BAD_MULTIPART: "OFPBRC_BAD_MULTIPART",
            ofp.OFPBRC_BAD_EXPERIMENTER: "OFPBRC_BAD_EXPERIMENTER",
            ofp.OFPBRC_BAD_EXP_TYPE: "OFPBRC_BAD_EXP_TYPE",
            ofp.OFPBRC_EPERM: "OFPBRC_EPERM",
            ofp.OFPBRC_BAD_LEN: "OFPBRC_BAD_LEN",
            ofp.OFPBRC_BUFFER_EMPTY: "OFPBRC_BUFFER_EMPTY",
            ofp.OFPBRC_BUFFER_UNKNOWN: "OFPBRC_BUFFER_UNKNOWN",
            ofp.OFPBRC_BAD_TABLE_ID: "OFPBRC_BAD_TABLE_ID",
            ofp.OFPBRC_IS_SLAVE: "OFPBRC_IS_SLAVE",
            ofp.OFPBRC_BAD_PORT: "OFPBRC_BAD_PORT",
            ofp.OFPBRC_BAD_PACKET: "OFPBRC_BAD_PACKET",
            ofp.OFPBRC_MULTIPART_BUFFER_OVERFLOW: "OFPBRC_MULTIPART_BUFFER_OVERFLOW",
        },
    ),
    ofp.OFPET_BAD_ACTION: (
        "OFPET_BAD_ACTION",
        {
            ofp.OFPBAC_BAD_TYPE: "OFPBAC_BAD_TYPE",
            ofp.OFPBAC_BAD_LEN: "OFPBAC_BAD_LEN",
            ofp.OFPBAC_BAD_EXPERIMENTER: "OFPBAC_BAD_EXPERIMENTER",
            ofp.OFPBAC_BAD_EXP_TYPE: "OFPBAC_BAD_EXP_TYPE",
            ofp.OFPBAC_BAD_OUT_PORT: "OFPBAC_BAD_OUT_PORT",
            ofp.OFPBAC_BAD_ARGUMENT: "OFPBAC_BAD_ARGUMENT",
            ofp.OFPBAC_EPERM: "OFPBAC_EPERM",
            ofp.OFPBAC_TOO_MANY: "OFPBAC_TOO_MANY",
            ofp.OFPBAC_BAD_QUEUE: "OFPBAC_BAD_QUEUE",
            ofp.OFPBAC_BAD_OUT_GROUP: "OFPBAC_BAD_OUT_GROUP",
            ofp.OFPBAC_MATCH_INCONSISTENT: "OFPBAC_MATCH_INCONSISTENT",
            ofp.OFPBAC_UNSUPPORTED_ORDER: "OFPBAC_UNSUPPORTED_ORDER",
            ofp.OFPBAC_BAD_TAG: "OFPBAC_BAD_TAG",
            ofp.OFPBAC_BAD_SET_TYPE: "OFPBAC_BAD_SET_TYPE",
            ofp.OFPBAC_BAD_SET_LEN: "OFPBAC_BAD_SET_LEN",
            ofp.OFPBAC_BAD_SET_ARGUMENT: "OFPBAC_BAD_SET_ARGUMENT",
        },
    ),
    ofp.OFPET_BAD_INSTRUCTION: (
        "OFPET_BAD_INSTRUCTION",
        {
            ofp.OFPBIC_UNKNOWN_INST: "OFPBIC_UNKNOWN_INST",
            ofp.OFPBIC_UNSUP_INST: "OFPBIC_UNSUP_INST",
            ofp.OFPBIC_BAD_TABLE_ID: "OFPBIC_BAD_TABLE_ID",
            ofp.OFPBIC_UNSUP_METADATA: "OFPBIC_UNSUP_METADATA",
            ofp.OFPBIC_UNSUP_METADATA_MASK: "OFPBIC_UNSUP_METADATA_MASK",
            ofp.OFPBIC_BAD_EXPERIMENTER: "OFPBIC_BAD_EXPERIMENTER",
            ofp.OFPBIC_BAD_EXP_TYPE: "OFPBIC_BAD_EXP_TYPE",
            ofp.OFPBIC_BAD_LEN: "OFPBIC_BAD_LEN",
            ofp.OFPBIC_EPERM: "OFPBIC_EPERM",
        },
    ),
    ofp.OFPET_BAD_MATCH: (
        "OFPET_BAD_MATCH",
        {
            ofp.OFPBMC_BAD_TYPE: "OFPBMC_BAD_TYPE",
            ofp.OFPBMC_BAD_LEN: "OFPBMC_BAD_LEN",
            ofp.OFPBMC_BAD_TAG: "OFPBMC_BAD_TAG",
            ofp.OFPBMC_BAD_DL_ADDR_MASK: "OFPBMC_BAD_DL_ADDR_MASK",
            ofp.OFPBMC_BAD_NW_ADDR_MASK: "OFPBMC_BAD_NW_ADDR_MASK",
            ofp.OFPBMC_BAD_WILDCARDS: "OFPBMC_BAD_WILDCARDS",
            ofp.OFPBMC_BAD_FIELD: "OFPBMC_BAD_FIELD",
            ofp.OFPBMC_BAD_VALUE: "OFPBMC_BAD_VALUE",
            ofp.OFPBMC_BAD_MASK: "OFPBMC_BAD_MASK",
            ofp.OFPBMC_BAD_PREREQ: "OFPBMC_BAD_PREREQ",
            ofp.OFPBMC_DUP_FIELD: "OFPBMC_DUP_FIELD",
            ofp.OFPBMC_EPERM: "OFPBMC_EPERM",
        },
    ),
    ofp.OFPET_FLOW_MOD_FAILED: (
        "OFPET_FLOW_MOD_FAILED",
        {
            ofp.OFPFMFC_UNKNOWN: "OFPFMFC_UNKNOWN",
            ofp.OFPFMFC_TABLE_FULL: "OFPFMFC_TABLE_FULL",
            ofp.OFPFMFC_BAD_TABLE_ID: "OFPFMFC_BAD_TABLE_ID",
            ofp.OFPFMFC_OVERLAP: "OFPFMFC_OVERLAP",
            ofp.OFPFMFC_EPERM: "OFPFMFC_EPERM",
            ofp.OFPFMFC_BAD_TIMEOUT: "OFPFMFC_BAD_TIMEOUT",
            ofp.OFPFMFC_BAD_COMMAND: "OFPFMFC_BAD_COMMAND",
            ofp.OFPFMFC_BAD_FLAGS: "OFPFMFC_BAD_FLAGS",
        },
    ),
    ofp.OFPET_GROUP_MOD_FAILED: (
        "OFPET_GROUP_MOD_FAILED",
        {
            ofp.OFPGMFC_GROUP_EXISTS: "OFPGMFC_GROUP_EXISTS",
            ofp.OFPGMFC_INVALID_GROUP: "OFPGMFC_INVALID_GROUP",
            ofp.OFPGMFC_WEIGHT_UNSUPPORTED: "OFPGMFC_WEIGHT_UNSUPPORTED",
            ofp.OFPGMFC_OUT_OF_GROUPS: "OFPGMFC_OUT_OF_GROUPS",
            ofp.OFPGMFC_OUT_OF_BUCKETS: "OFPGMFC_OUT_OF_BUCKETS",
            ofp.OFPGMFC_CHAINING_UNSUPPORTED: "OFPGMFC_CHAINING_UNSUPPORTED",
            ofp.OFPGMFC_WATCH_UNSUPPORTED: "OFPGMFC_WATCH_UNSUPPORTED",
            ofp.OFPGMFC_LOOP: "OFPGMFC_LOOP",
            ofp.OFPGMFC_UNKNOWN_GROUP: "OFPGMFC_UNKNOWN_GROUP",
            ofp.OFPGMFC_CHAINED_GROUP: "OFPGMFC_CHAINED_GROUP",
            ofp.OFPGMFC_BAD_TYPE: "OFPGMFC_BAD_TYPE",
            ofp.OFPGMFC_BAD_COMMAND: "OFPGMFC_BAD_COMMAND",
            ofp.OFPGMFC_BAD_BUCKET: "OFPGMFC_BAD_BUCKET",
            ofp.OFPGMFC_BAD_WATCH: "OFPGMFC_BAD_WATCH",
            ofp.OFPGMFC_EPERM: "OFPGMFC_EPERM",
        },
    ),
    ofp.OFPET_PORT_MOD_FAILED: (
        "OFPET_PORT_MOD_FAILED",
        {
            ofp.OFPPMFC_BAD_PORT: "OFPPMFC_BAD_PORT",
            ofp.OFPPMFC_BAD_HW_ADDR: "OFPPMFC_BAD_HW_ADDR",
            ofp.OFPPMFC_BAD_CONFIG: "OFPPMFC_BAD_CONFIG",
            ofp.OFPPMFC_BAD_ADVERTISE: "OFPPMFC_BAD_ADVERTISE",
            ofp.OFPPMFC_EPERM: "OFPPMFC_EPERM",
        },
    ),
    ofp.OFPET_TABLE_MOD_FAILED: (
        "OFPET_TABLE_MOD_FAILED",
        {
            ofp.OFPTMFC_BAD_TABLE: "OFPTMFC_BAD_TABLE",
            ofp.OFPTMFC_BAD_CONFIG: "OFPTMFC_BAD_CONFIG",
            ofp.OFPTMFC_EPERM: "OFPTMFC_EPERM",
        },
    ),
    ofp.OFPET_QUEUE_OP_FAILED: (
        "OFPET_QUEUE_OP_FAILED",
        {
            ofp.OFPQOFC_BAD_PORT: "OFPQOFC_BAD_PORT",
            ofp.OFPQOFC_BAD_QUEUE: "OFPQOFC_BAD_QUEUE",
            ofp.OFPQOFC_EPERM: "OFPQOFC_EPERM",
        },
    ),
    ofp.OFPET_SWITCH_CONFIG_FAILED: (
        "OFPET_SWITCH_CONFIG_FAILED",
        {
            ofp.OFPSCFC_BAD_FLAGS: "OFPSCFC_BAD_FLAGS",
            ofp.OFPSCFC_BAD_LEN: "OFPSCFC_BAD_LEN",
            ofp.OFPSCFC_EPERM: "OFPSCFC_EPERM",
        },
    ),
    ofp.OFPET_ROLE_REQUEST_FAILED: (
        "OFPET_ROLE_REQUEST_FAILED",
        {
            ofp.OFPRRFC_STALE: "OFPRRFC_STALE",
            ofp.OFPRRFC_UNSUP: "OFPRRFC_UNSUP",
            ofp.OFPRRFC_BAD_ROLE: "OFPRRFC_BAD_ROLE",
        },
    ),
    ofp.OFPET_METER_MOD_FAILED: (
        "OFPET_METER_MOD_FAILED",
        {
            ofp.OFPMMFC_UNKNOWN: "OFPMMFC_UNKNOWN",
            ofp.OFPMMFC_METER_EXISTS: "OFPMMFC_METER_EXISTS",
            ofp.OFPMMFC_INVALID_METER: "OFPMMFC_INVALID_METER",
            ofp.OFPMMFC_UNKNOWN_METER: "OFPMMFC_UNKNOWN_METER",
            ofp.OFPMMFC_BAD_COMMAND: "OFPMMFC_BAD_COMMAND",
            ofp.OFPMMFC_BAD_FLAGS: "OFPMMFC_BAD_FLAGS",
            ofp.OFPMMFC_BAD_RATE: "OFPMMFC_BAD_RATE",
            ofp.OFPMMFC_BAD_BURST: "OFPMMFC_BAD_BURST",
            ofp.OFPMMFC_BAD_BAND: "OFPMMFC_BAD_BAND",
            ofp.OFPMMFC_BAD_BAND_VALUE: "OFPMMFC_BAD_BAND_VALUE",
            ofp.OFPMMFC_OUT_OF_METERS: "OFPMMFC_OUT_OF_METERS",
            ofp.OFPMMFC_OUT_OF_BANDS: "OFPMMFC_OUT_OF_BANDS",
        },
    ),
    ofp.OFPET_TABLE_FEATURES_FAILED: (
        "OFPET_TABLE_FEATURES_FAILED",
        {
            ofp.OFPTFFC_BAD_TABLE: "OFPTFFC_BAD_TABLE",
            ofp.OFPTFFC_BAD_METADATA: "OFPTFFC_BAD_METADATA",
            ofp.OFPTFFC_BAD_TYPE: "OFPTFFC_BAD_TYPE",
            ofp.OFPTFFC_BAD_LEN: "OFPTFFC_BAD_LEN",
            ofp.OFPTFFC_BAD_ARGUMENT: "OFPTFFC_BAD_ARGUMENT",
            ofp.OFPTFFC_EPERM: "OFPTFFC_EPERM",
        },
    ),
    ofp.OFPET_EXPERIMENTER: ("OFPET_EXPERIMENTER", {}),
}


def ignore_port(port_num):
    """Return True if FAUCET should ignore this port.

    Args:
        port_num (int): switch port.
    Returns:
        bool: True if FAUCET should ignore this port.
    """
    # special case OFPP_LOCAL to allow FAUCET to manage switch admin interface.
    if port_num in (ofp.OFPP_LOCAL, ofp.OFPP_IN_PORT):
        return False
    # 0xF0000000 and up are not physical ports.
    return port_num > 0xF0000000


def port_status_from_state(state):
    """Return True if OFPPS_LINK_DOWN is not set."""
    return not state & ofp.OFPPS_LINK_DOWN


def is_table_features_req(ofmsg):
    """Return True if flow message is a TFM req.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a TFM req.
    """
    return isinstance(ofmsg, parser.OFPTableFeaturesStatsRequest)


def is_flowmod(ofmsg):
    """Return True if flow message is a FlowMod.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a FlowMod
    """
    return isinstance(ofmsg, parser.OFPFlowMod)


def is_flowaddmod(ofmsg):
    """Return True if flow message is a FlowMod, add or modify.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a FlowMod, add or modify.
    """
    return isinstance(ofmsg, parser.OFPFlowMod) and ofmsg.command in (
        ofp.OFPFC_ADD,
        ofp.OFPFC_MODIFY,
        ofp.OFPFC_MODIFY_STRICT,
    )


def is_groupmod(ofmsg):
    """Return True if OF message is a GroupMod.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a GroupMod
    """
    return isinstance(ofmsg, parser.OFPGroupMod)


def is_metermod(ofmsg):
    """Return True if OF message is a MeterMod.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a MeterMod
    """
    return isinstance(ofmsg, parser.OFPMeterMod)


def is_packetout(ofmsg):
    """Return True if OF message is a PacketOut

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a PacketOut
    """
    return isinstance(ofmsg, parser.OFPPacketOut)


def is_output(ofmsg):
    """Return True if flow message is an action output message.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a OFPActionOutput.
    """
    return isinstance(ofmsg, parser.OFPActionOutput)


def is_flowdel(ofmsg):
    """Return True if flow message is a FlowMod and a delete.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a FlowMod delete/strict.
    """
    return is_flowmod(ofmsg) and ofmsg.command in (
        ofp.OFPFC_DELETE,
        ofp.OFPFC_DELETE_STRICT,
    )


def is_groupdel(ofmsg):
    """Return True if OF message is a GroupMod and command is delete.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a GroupMod delete
    """
    if is_groupmod(ofmsg) and (ofmsg.command == ofp.OFPGC_DELETE):
        return True
    return False


def is_meterdel(ofmsg):
    """Return True if OF message is a MeterMod and command is delete.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a MeterMod delete
    """
    if is_metermod(ofmsg) and (ofmsg.command == ofp.OFPMC_DELETE):
        return True
    return False


def is_groupadd(ofmsg):
    """Return True if OF message is a GroupMod and command is add.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a GroupMod add
    """
    if is_groupmod(ofmsg) and (ofmsg.command == ofp.OFPGC_ADD):
        return True
    return False


def is_meteradd(ofmsg):
    """Return True if OF message is a MeterMod and command is add.

    Args:
        ofmsg: ryu.ofproto.ofproto_v1_3_parser message.
    Returns:
        bool: True if is a MeterMod add
    """
    if is_metermod(ofmsg) and (ofmsg.command == ofp.OFPMC_ADD):
        return True
    return False


def is_apply_actions(instruction):
    """Return True if an apply action.

    Args:
        instruction: OpenFlow instruction.
    Returns:
        bool: True if an apply action.
    """
    return (
        isinstance(instruction, parser.OFPInstructionActions)
        and instruction.type == ofp.OFPIT_APPLY_ACTIONS
    )


def is_meter(instruction):
    """Return True if a meter.

    Args:
        instruction: OpenFlow instruction.
    Returns:
        bool: True if a meter.
    """
    return isinstance(instruction, parser.OFPInstructionMeter)


def is_set_field(action):
    return isinstance(action, parser.OFPActionSetField)


def is_ct(action):
    return isinstance(action, parser.NXActionCT)  # pylint: disable=no-member


def apply_meter(meter_id):
    """Return instruction to apply a meter."""
    return parser.OFPInstructionMeter(meter_id, ofp.OFPIT_METER)


@functools.lru_cache()
def _apply_actions(actions):
    return parser.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, actions)


def apply_actions(actions):
    """Return instruction that applies action list.

    Args:
        actions (list): list of OpenFlow actions.
    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPInstruction: instruction of actions.
    """
    return _apply_actions(tuple(actions))


@functools.lru_cache()
def goto_table(table):
    """Return instruction to goto table.

    Args:
        table (ValveTable): table to goto.
    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPInstruction: goto instruction.
    """
    return parser.OFPInstructionGotoTable(table.table_id)


@functools.lru_cache()
def goto_table_id(table_id):
    """Return instruction to goto table by table ID.

    Args:
        table (int): table by ID to goto.
    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPInstruction: goto instruction.
    """
    return parser.OFPInstructionGotoTable(table_id)


def metadata_goto_table(metadata, mask, table):
    """Return instructions to write metadata and goto table.

    Args:
        metadata (int): metadata to write to packet
        maks (int): mask to apply to metadata
        table (ValveTable): table to goto.
    Returns:
        list of OFPInstructions"""
    return [
        parser.OFPInstructionWriteMetadata(metadata, mask),
        parser.OFPInstructionGotoTable(table.table_id),
    ]


@functools.lru_cache()
def set_field(**kwds):
    """Return action to set any field.

    Args:
        kwds (dict): exactly one field to set
    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPActionSetField: set field action.
    """
    return parser.OFPActionSetField(**kwds)


def vid_present(vid):
    """Return VLAN VID with VID_PRESENT flag set.

    Args:
        vid (int): VLAN VID
    Returns:
        int: VLAN VID with VID_PRESENT.
    """
    return vid | ofp.OFPVID_PRESENT


def devid_present(vid):
    """Return VLAN VID without VID_PRESENT flag set.

    Args:
        vid (int): VLAN VID with VID_PRESENT.
    Returns:
        int: VLAN VID.
    """
    return vid ^ ofp.OFPVID_PRESENT


@functools.lru_cache(maxsize=1024)
def push_vlan_act(table, vlan_vid, eth_type=ether.ETH_TYPE_8021Q):
    """Return OpenFlow action list to push Ethernet 802.1Q header with VLAN VID.

    Args:
        vid (int): VLAN VID
    Returns:
        list: actions to push 802.1Q header with VLAN VID set.
    """
    return [
        parser.OFPActionPushVlan(eth_type),
        table.set_vlan_vid(vlan_vid),
    ]


@functools.lru_cache()
def dec_ip_ttl():
    """Return OpenFlow action to decrement IP TTL.

    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPActionDecNwTtl: decrement IP TTL.
    """
    return parser.OFPActionDecNwTtl()


@functools.lru_cache(maxsize=1024)
def pop_vlan():
    """Return OpenFlow action to pop outermost Ethernet 802.1Q VLAN header.

    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPActionPopVlan: Pop VLAN.
    """
    return parser.OFPActionPopVlan()


def ct(**kwds):  # pylint: disable=invalid-name
    """Return connection tracker action.

    Args:
        kwds (dict): exactly one connection tracker action.
    Returns:
        ryu.ofproto.nx_actions.NXActionCT: connection tracker action.
    """
    return parser.NXActionCT(**kwds)  # pylint: disable=no-member


def ct_clear():
    """Return clear connection tracker state action.

    Args:
        kwds (dict): exactly one clear connection tracker state action.
    Returns:
        ryu.ofproto.nx_actions.NXActionCTClear: clear connection tracker state action.
    """
    return parser.NXActionCTClear()  # pylint: disable=no-member


def ct_nat(**kwds):
    """Return network address translation connection tracker action.

    Args:
        kwds (dict): exactly one network address translation connection tracker action.
    Returns:
        ryu.ofproto.nx_actions.NXActionNAT: network address translation connection tracker action.
    """
    return parser.NXActionNAT(**kwds)  # pylint: disable=no-member


@functools.lru_cache(maxsize=1024)
def output_port(port_num, max_len=0):
    """Return OpenFlow action to output to a port.

    Args:
        port_num (int): port to output to.
        max_len (int): maximum length of packet to output (default no maximum).
    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput: output to port action.
    """
    return parser.OFPActionOutput(port_num, max_len=max_len)


def ports_from_output_port_acts(output_port_acts):
    """Return unique port numbers from OFPActionOutput actions.

    Args:
        list of ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput: output to port actions.
    Returns:
        set of port number ints.
    """
    return {output_port_act.port for output_port_act in output_port_acts}


def dedupe_output_port_acts(output_port_acts):
    """Deduplicate parser.OFPActionOutputs (because Ryu doesn't define __eq__).

    Args:
        list of ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput: output to port actions.
    Returns:
        list of ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput: output to port actions.
    """
    output_ports = ports_from_output_port_acts(output_port_acts)
    return [output_port(port) for port in sorted(output_ports)]


@functools.lru_cache(maxsize=1024)
def output_non_output_actions(flood_acts):
    """Split output actions into deduped actions, output ports, and non-output port actions.

    Args:
        list of ryu.ofproto.ofproto_v1_3_parser.OFPActions: flood actions.
    Returns:
        set of deduped actions, output ports, and non-output actions.
    """
    output_ports = set()
    all_nonoutput_actions = set()
    deduped_acts = []
    # avoid dedupe_ofmsgs() here, as it's expensive - most of the time we are comparing
    # port numbers as integers which is much cheaper.
    for act in flood_acts:
        if is_output(act):
            if act.port in output_ports:
                continue
            output_ports.add(act.port)
        else:
            str_act = str(act)
            if str_act in all_nonoutput_actions:
                continue
            all_nonoutput_actions.add(str_act)
        deduped_acts.append(act)
    nonoutput_actions = all_nonoutput_actions - set([str(pop_vlan())])
    return (deduped_acts, output_ports, nonoutput_actions)


@functools.lru_cache()
def output_in_port():
    """Return OpenFlow action to output out input port.

    Returns:
       ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput.
    """
    return output_port(OFP_IN_PORT)


@functools.lru_cache()
def output_controller(max_len=MAX_PACKET_IN_BYTES):
    """Return OpenFlow action to packet in to the controller.

    Args:
        max_len (int): max number of bytes from packet to output.
    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput: packet in action.
    """
    return output_port(ofp.OFPP_CONTROLLER, max_len)


def packetouts(port_nums, data):
    """Return OpenFlow action to multiply packet out to dataplane from controller.

    Args:
        port_num (list): ints, ports to output to.
        data (str): raw packet to output.
    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput: packet out action.
    """
    random.shuffle(port_nums)
    return parser.OFPPacketOut(
        datapath=None,
        buffer_id=ofp.OFP_NO_BUFFER,
        in_port=ofp.OFPP_CONTROLLER,
        actions=[output_port(port_num) for port_num in port_nums],
        data=data,
    )


@functools.lru_cache()
def packetout(port_num, data):
    """Return OpenFlow action to packet out to dataplane from controller.

    Args:
        port_num (int): port to output to.
        data (str): raw packet to output.
    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPActionOutput: packet out action.
    """
    return packetouts([port_num], data)


@functools.lru_cache()
def barrier():
    """Return OpenFlow barrier request.

    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPBarrierRequest: barrier request.
    """
    return parser.OFPBarrierRequest(None)


def table_features(body):
    return parser.OFPTableFeaturesStatsRequest(datapath=None, body=body)


def match(match_fields):
    """Return OpenFlow matches from dict.

    Args:
        match_fields (dict): match fields and values.
    Returns:
        ryu.ofproto.ofproto_v1_3_parser.OFPMatch: matches.
    """
    return parser.OFPMatch(**match_fields)


@functools.lru_cache()
def valve_match_vid(value):
    return to_match_vid(value, ofp.OFPVID_PRESENT)


MATCH_FIELDS = {
    # See 7.2.3.7 Flow Match Fields (OF 1.3.5)
    "in_port": OFCtlUtil(ofp).ofp_port_from_user,
    "in_phy_port": str_to_int,
    "metadata": to_match_masked_int,
    "eth_dst": to_match_eth,
    "eth_src": to_match_eth,
    "eth_type": str_to_int,
    "vlan_vid": valve_match_vid,
    "vlan_pcp": str_to_int,
    "ip_dscp": str_to_int,
    "ip_ecn": str_to_int,
    "ip_proto": str_to_int,
    "ipv4_src": to_match_ip,
    "ipv4_dst": to_match_ip,
    "tcp_src": to_match_masked_int,
    "tcp_dst": to_match_masked_int,
    "udp_src": to_match_masked_int,
    "udp_dst": to_match_masked_int,
    "sctp_src": to_match_masked_int,
    "sctp_dst": to_match_masked_int,
    "icmpv4_type": str_to_int,
    "icmpv4_code": str_to_int,
    "arp_op": str_to_int,
    "arp_spa": to_match_ip,
    "arp_tpa": to_match_ip,
    "arp_sha": to_match_eth,
    "arp_tha": to_match_eth,
    "ipv6_src": to_match_ip,
    "ipv6_dst": to_match_ip,
    "ipv6_flabel": str_to_int,
    "icmpv6_type": str_to_int,
    "icmpv6_code": str_to_int,
    "ipv6_nd_target": to_match_ip,
    "ipv6_nd_sll": to_match_eth,
    "ipv6_nd_tll": to_match_eth,
    "mpls_label": str_to_int,
    "mpls_tc": str_to_int,
    "mpls_bos": str_to_int,
    "pbb_isid": to_match_masked_int,
    "tunnel_id": to_match_masked_int,
    "ipv6_exthdr": to_match_masked_int,
    # Nicira extensions, see ovs-fields(7)
    "ct_state": to_match_masked_int,
    "ct_zone": str_to_int,
    "ct_mark": to_match_masked_int,
    "ct_label": to_match_masked_int,
}


def match_from_dict(match_dict):
    """Parse a match dict into a OFPMatch object"""
    kwargs = {}
    for of_match, field in match_dict.items():
        of_match = OLD_MATCH_FIELDS.get(of_match, of_match)
        test_config_condition(
            of_match not in MATCH_FIELDS, "Unknown match field: %s" % of_match
        )
        try:
            encoded_field = MATCH_FIELDS[of_match](field)
        except TypeError as type_error:
            raise InvalidConfigError(
                "%s cannot be type %s" % (of_match, type(field))
            ) from type_error
        kwargs[of_match] = encoded_field

    return parser.OFPMatch(**kwargs)


def _match_ip_masked(ipa):
    if isinstance(ipa, (ipaddress.IPv4Network, ipaddress.IPv6Network)):
        return (str(ipa.network_address), str(ipa.netmask))
    return (str(ipa.ip), str(ipa.netmask))


@functools.lru_cache(maxsize=1024)
def build_match_dict(
    in_port=None,
    vlan=None,
    eth_type=None,
    eth_src=None,
    eth_dst=None,
    eth_dst_mask=None,
    icmpv6_type=None,
    nw_proto=None,
    nw_dst=None,
    metadata=None,
    metadata_mask=None,
    vlan_pcp=None,
    udp_src=None,
    udp_dst=None,
):
    match_dict = {}
    if in_port is not None:
        match_dict["in_port"] = in_port
    if vlan is not None:
        if isinstance(vlan, int):
            vid = vlan
        else:
            vid = vlan.vid
        if vid == ofp.OFPVID_NONE:
            match_dict["vlan_vid"] = int(ofp.OFPVID_NONE)
        elif vid == ofp.OFPVID_PRESENT:
            match_dict["vlan_vid"] = (ofp.OFPVID_PRESENT, ofp.OFPVID_PRESENT)
        else:
            match_dict["vlan_vid"] = vid_present(vid)
    if eth_src is not None:
        match_dict["eth_src"] = eth_src
    if eth_dst is not None:
        if eth_dst_mask is not None:
            match_dict["eth_dst"] = (eth_dst, eth_dst_mask)
        else:
            match_dict["eth_dst"] = eth_dst
    if nw_proto is not None:
        match_dict["ip_proto"] = nw_proto
    if udp_dst is not None:
        match_dict["udp_dst"] = udp_dst
    if udp_src is not None:
        match_dict["udp_src"] = udp_src
    if icmpv6_type is not None:
        match_dict["icmpv6_type"] = icmpv6_type
    if nw_dst is not None:
        nw_dst_masked = _match_ip_masked(nw_dst)
        if eth_type == ether.ETH_TYPE_ARP:
            match_dict["arp_tpa"] = str(nw_dst.ip)
        elif eth_type == ether.ETH_TYPE_IP:
            match_dict["ipv4_dst"] = nw_dst_masked
        else:
            match_dict["ipv6_dst"] = nw_dst_masked
    if eth_type is not None:
        match_dict["eth_type"] = eth_type
    if metadata is not None:
        if metadata_mask is not None:
            match_dict["metadata"] = (metadata, metadata_mask)
        else:
            match_dict["metadata"] = metadata
    if vlan_pcp is not None:
        match_dict["vlan_pcp"] = vlan_pcp
    return match_dict


@functools.lru_cache()
def flowmod(
    cookie,
    command,
    table_id,
    priority,
    out_port,
    out_group,
    match_fields,
    inst,
    hard_timeout,
    idle_timeout,
    flags=0,
):
    return parser.OFPFlowMod(
        datapath=None,
        cookie=cookie,
        command=command,
        table_id=table_id,
        priority=priority,
        out_port=out_port,
        out_group=out_group,
        match=match_fields,
        instructions=inst,
        hard_timeout=hard_timeout,
        idle_timeout=idle_timeout,
        flags=flags,
    )


class NullRyuDatapath:
    """Placeholder Ryu Datapath."""

    ofproto = ofp


@functools.lru_cache()
def verify_flowmod(flowmod_msg):
    """Verify flowmod can be serialized."""
    flowmod_msg.datapath = NullRyuDatapath()
    # Must be non-zero.
    flowmod_msg.set_xid(1)
    flowmod_msg.serialize()


def group_act(group_id):
    """Return an action to run a group."""
    return parser.OFPActionGroup(group_id)


def bucket(weight=0, watch_port=ofp.OFPP_ANY, watch_group=ofp.OFPG_ANY, actions=None):
    """Return a group action bucket with provided actions."""
    return parser.OFPBucket(
        weight=weight, watch_port=watch_port, watch_group=watch_group, actions=actions
    )


def build_group_flood_buckets(vlan_flood_acts):
    """Return a list of group buckets to implement flooding on a VLAN."""
    buckets = []
    non_outputs = []
    for act in vlan_flood_acts:
        if is_output(act):
            buckets.append(bucket(actions=non_outputs + [act]))
        else:
            non_outputs.append(act)
    return buckets


def groupdel(datapath=None, group_id=ofp.OFPG_ALL):
    """Delete a group (default all groups)."""
    return parser.OFPGroupMod(datapath, ofp.OFPGC_DELETE, 0, group_id)


def groupadd(datapath=None, type_=ofp.OFPGT_ALL, group_id=0, buckets=None):
    """Add a group."""
    return [
        groupdel(datapath=datapath, group_id=group_id),
        parser.OFPGroupMod(datapath, ofp.OFPGC_ADD, type_, group_id, buckets),
    ]


def groupadd_ff(datapath=None, group_id=0, buckets=None):
    """Add a fast failover group."""
    return groupadd(datapath, type_=ofp.OFPGT_FF, group_id=group_id, buckets=buckets)


def meterdel(datapath=None, meter_id=ofp.OFPM_ALL):
    """Delete a meter (default all meters)."""
    return parser.OFPMeterMod(datapath, ofp.OFPMC_DELETE, 0, meter_id)


def meteradd(meter_conf, command=ofp.OFPMC_ADD):
    """Add a meter based on YAML configuration."""

    class NoopDP:
        """Fake DP to be able to use ofctl to parse meter config."""

        id = 0
        msg = None
        ofproto = ofp
        ofproto_parser = parser

        def send_msg(self, msg):
            """Save msg only."""
            self.msg = msg

        @staticmethod
        def set_xid(msg):
            """Clear msg XID."""
            msg.xid = 0

    noop_dp = NoopDP()
    ofctl.mod_meter_entry(noop_dp, meter_conf, command)
    noop_dp.msg.xid = None
    noop_dp.msg.datapath = None
    return noop_dp.msg


def controller_pps_meteradd(datapath=None, pps=0):
    """Add a PPS meter towards controller."""
    return parser.OFPMeterMod(
        datapath=datapath,
        command=ofp.OFPMC_ADD,
        flags=ofp.OFPMF_PKTPS,
        meter_id=ofp.OFPM_CONTROLLER,
        bands=[parser.OFPMeterBandDrop(rate=pps)],
    )


def controller_pps_meterdel(datapath=None):
    """Delete a PPS meter towards controller."""
    return parser.OFPMeterMod(
        datapath=datapath,
        command=ofp.OFPMC_DELETE,
        flags=ofp.OFPMF_PKTPS,
        meter_id=ofp.OFPM_CONTROLLER,
    )


def slowpath_pps_meteradd(datapath=None, pps=0):
    """Add a PPS meter towards controller."""
    return parser.OFPMeterMod(
        datapath=datapath,
        command=ofp.OFPMC_ADD,
        flags=ofp.OFPMF_PKTPS,
        meter_id=ofp.OFPM_SLOWPATH,
        bands=[parser.OFPMeterBandDrop(rate=pps)],
    )


def slowpath_pps_meterdel(datapath=None):
    """Delete a PPS meter towards controller."""
    return parser.OFPMeterMod(
        datapath=datapath,
        command=ofp.OFPMC_DELETE,
        flags=ofp.OFPMF_PKTPS,
        meter_id=ofp.OFPM_SLOWPATH,
    )


def is_global_flowdel(ofmsg):
    """Is a delete of all flows in all tables."""
    return (
        is_flowdel(ofmsg)
        and ofmsg.table_id == ofp.OFPTT_ALL
        and not ofmsg.match.items()
    )


def is_global_groupdel(ofmsg):
    """Is a delete of all groups."""
    return is_groupdel(ofmsg) and ofmsg.group_id == ofp.OFPG_ALL


def is_global_meterdel(ofmsg):
    """Is a delete of all meters."""
    return is_meterdel(ofmsg) and ofmsg.meter_id == ofp.OFPM_ALL


# We can tell right away what kind of OF messages these are.
_MSG_KINDS_TYPES = {
    parser.OFPPacketOut: "packetout",
    parser.OFPTableFeaturesStatsRequest: "tfm",
    parser.OFPSetConfig: "config",
    parser.OFPSetAsync: "config",
    parser.OFPDescStatsRequest: "config",
}


# We need to examine the OF message more closely to classify it.
_MSG_KINDS = {
    parser.OFPFlowMod: (
        ("deleteglobal", is_global_flowdel),
        ("delete", is_flowdel),
        ("flowaddmod", is_flowaddmod),
    ),
    parser.OFPGroupMod: (
        ("deleteglobal", is_global_groupdel),
        ("delete", is_groupdel),
        ("groupadd", is_groupadd),
    ),
    parser.OFPMeterMod: (
        ("deleteglobal", is_global_meterdel),
        ("delete", is_meterdel),
        ("meteradd", is_meteradd),
    ),
}


@functools.lru_cache()
def _msg_kind(ofmsg):
    ofmsg_type = type(ofmsg)
    ofmsg_kind = _MSG_KINDS_TYPES.get(ofmsg_type, None)
    if ofmsg_kind:
        return ofmsg_kind
    kinds = _MSG_KINDS.get(ofmsg_type, None)
    if kinds:
        for kind, kind_func in kinds:
            if kind_func(ofmsg):
                return kind
    return "other"


def _partition_ofmsgs(input_ofmsgs):
    """Partition input ofmsgs by kind."""
    by_kind = {}
    for ofmsg in input_ofmsgs:
        by_kind.setdefault(_msg_kind(ofmsg), []).append(ofmsg)
    return by_kind


def _flowmodkey(ofmsg):
    return (ofmsg.match, ofmsg.cookie, ofmsg.priority, ofmsg.table_id)


def _none_flowmodkey(ofmsg):
    try:
        return _flowmodkey(ofmsg)
    except AttributeError:
        return None


def sort_flows(input_ofmsgs):
    """Sort flows in canonical order, descending table and priority."""
    return sorted(
        input_ofmsgs,
        key=lambda ofmsg: (
            getattr(ofmsg, "table_id", ofp.OFPTT_ALL),
            getattr(ofmsg, "priority", 2**16 + 1),
        ),
        reverse=True,
    )


def dedupe_ofmsgs(input_ofmsgs, random_order, flowkey):
    """Return deduplicated ofmsg list."""
    # Built in comparison doesn't work until serialized() called
    # Can't use dict or json comparison as may be nested
    deduped_input_ofmsgs = {flowkey(ofmsg): ofmsg for ofmsg in input_ofmsgs}
    if random_order:
        ofmsgs = list(deduped_input_ofmsgs.values())
        random.shuffle(ofmsgs)
        return ofmsgs
    return sort_flows(deduped_input_ofmsgs.values())


def dedupe_overlaps_ofmsgs(input_ofmsgs, random_order, flowkey):
    deduped_ofmsgs = dedupe_ofmsgs(input_ofmsgs, random_order, flowkey)
    ofmsgs_by_table = {}
    for ofmsg in deduped_ofmsgs:
        table_id = getattr(ofmsg, "table_id", None)
        ofmsgs_by_table.setdefault(table_id, []).append(ofmsg)
    all_table_ids = {
        table_id for table_id in ofmsgs_by_table if isinstance(table_id, int)
    }

    # If priority-less deletes across all tables are detected, then remove any
    # overlapping deletes (e.g. if a delete all tables vlan=100 is deleted, then remove
    # all other table-specific deletes that have vlan=100).
    if ofp.OFPTT_ALL in all_table_ids:
        overlap_matches = {
            tuple(ofmsg.match.items())
            for ofmsg in ofmsgs_by_table[ofp.OFPTT_ALL]
            if not ofmsg.priority
        }
        table_ids = all_table_ids - {ofp.OFPTT_ALL}
        if overlap_matches and table_ids:
            for table_id in table_ids:
                for overlap_match in overlap_matches:
                    overlap_match = set(overlap_match)
                    ofmsgs_by_table[table_id] = [
                        ofmsg
                        for ofmsg in ofmsgs_by_table[table_id]
                        if not overlap_match.issubset(set(ofmsg.match.items()))
                    ]
            nooverlaps_ofmsgs = []
            for _, ofmsgs in sorted(ofmsgs_by_table.items(), reverse=True):
                nooverlaps_ofmsgs.extend(ofmsgs)
            return nooverlaps_ofmsgs

    return deduped_ofmsgs


def remove_overlap_ofmsgs(input_ofmsgs, overlap_input_ofmsgs):
    overlap_keys = {_none_flowmodkey(ofmsg) for ofmsg in overlap_input_ofmsgs} - {None}
    input_ofmsgs_without_overlaps = []
    for ofmsg in input_ofmsgs:
        key = _none_flowmodkey(ofmsg)
        if key is not None and key in overlap_keys:
            continue
        input_ofmsgs_without_overlaps.append(ofmsg)
    return input_ofmsgs_without_overlaps


# kind, random_order, suggest_barrier, flowkey
_OFMSG_ORDER = (
    ("config", False, True, str, dedupe_ofmsgs),
    ("deleteglobal", False, True, str, dedupe_ofmsgs),
    ("delete", False, True, str, dedupe_overlaps_ofmsgs),
    ("tfm", False, True, str, dedupe_ofmsgs),
    ("groupadd", False, True, str, dedupe_ofmsgs),
    ("meteradd", False, True, str, dedupe_ofmsgs),
    ("flowaddmod", False, False, _flowmodkey, dedupe_ofmsgs),
    ("other", False, False, str, dedupe_ofmsgs),
    ("packetout", True, False, str, dedupe_ofmsgs),
)


def valve_flowreorder(input_ofmsgs, use_barriers=True):
    """Reorder flows for better OFA performance."""
    # Move all deletes to be first, and add one barrier,
    # while optionally randomizing order. Platforms that do
    # parallel delete will perform better and platforms that
    # don't will have at most only one barrier to deal with.
    output_ofmsgs = []
    by_kind = _partition_ofmsgs(input_ofmsgs)

    # Suppress all other relevant deletes if a global delete is present.
    delete_global_ofmsgs = by_kind.get("deleteglobal", [])
    if delete_global_ofmsgs:
        global_types = {type(ofmsg) for ofmsg in delete_global_ofmsgs}
        new_delete = [
            ofmsg
            for ofmsg in by_kind.get("delete", [])
            if type(ofmsg) not in global_types
        ]
        by_kind["delete"] = new_delete

    for kind, random_order, _suggest_barrier, flowkey, dedupe_func in _OFMSG_ORDER:
        ofmsgs = dedupe_func(by_kind.get(kind, []), random_order, flowkey)
        if ofmsgs:
            by_kind[kind] = ofmsgs

    deletes = by_kind.get("delete", None)
    addmod = by_kind.get("flowaddmod", None)
    if deletes and addmod:
        by_kind["delete"] = remove_overlap_ofmsgs(deletes, addmod)

    for kind, _random_order, suggest_barrier, _flowkey, dedupe_func in _OFMSG_ORDER:
        ofmsgs = by_kind.get(kind, [])
        if ofmsgs:
            output_ofmsgs.extend(ofmsgs)
            if use_barriers and suggest_barrier:
                output_ofmsgs.append(barrier())
    return output_ofmsgs


def flood_tagged_port_outputs(ports, in_port=None, exclude_ports=None):
    """Return list of actions necessary to flood to list of tagged ports."""
    flood_acts = []
    in_port_mirror_output_ports = {}
    if in_port is not None:
        in_port_mirror_output_ports = ports_from_output_port_acts(
            in_port.mirror_actions()
        )
    if ports:
        for port in ports:
            if in_port is not None and port == in_port:
                if in_port.hairpin:
                    flood_acts.append(output_in_port())
                continue
            if exclude_ports and port in exclude_ports:
                continue
            flood_acts.append(output_port(port.number))
            # Only mirror if different mirror actions to in_port
            # (already will be mirrored on input).
            mirror_actions = port.mirror_actions()
            mirror_output_ports = ports_from_output_port_acts(mirror_actions)
            if in_port is None or in_port_mirror_output_ports != mirror_output_ports:
                flood_acts.extend(mirror_actions)
    return dedupe_output_port_acts(flood_acts)


def flood_untagged_port_outputs(ports, in_port=None, exclude_ports=None):
    """Return list of actions necessary to flood to list of untagged ports."""
    flood_acts = flood_tagged_port_outputs(
        ports, in_port=in_port, exclude_ports=exclude_ports
    )
    if flood_acts:
        flood_acts = [pop_vlan()] + flood_acts
    return flood_acts


def flood_port_outputs(tagged_ports, untagged_ports, in_port=None, exclude_ports=None):
    """Return actions for both tagged and untagged ports."""
    return flood_tagged_port_outputs(
        tagged_ports, in_port, exclude_ports
    ) + flood_untagged_port_outputs(untagged_ports, in_port, exclude_ports)


def faucet_config(datapath=None):
    """Return switch config for FAUCET."""
    return parser.OFPSetConfig(datapath, ofp.OFPC_FRAG_NORMAL, 0)


def faucet_async(
    datapath=None, notify_flow_removed=False, packet_in=True, port_status=True
):
    """Return async message config for FAUCET/Gauge"""
    packet_in_mask = 0
    if packet_in:
        packet_in_mask = 1 << ofp.OFPR_ACTION
    port_status_mask = 0
    if port_status:
        port_status_mask = (
            1 << ofp.OFPPR_ADD | 1 << ofp.OFPPR_DELETE | 1 << ofp.OFPPR_MODIFY
        )
    flow_removed_mask = 0
    if notify_flow_removed:
        flow_removed_mask = 1 << ofp.OFPRR_IDLE_TIMEOUT | 1 << ofp.OFPRR_HARD_TIMEOUT
    return parser.OFPSetAsync(
        datapath,
        [packet_in_mask, packet_in_mask],
        [port_status_mask, port_status_mask],
        [flow_removed_mask, flow_removed_mask],
    )


def desc_stats_request(datapath=None):
    """Query switch description."""
    return parser.OFPDescStatsRequest(datapath, 0)
